﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Runtime.Serialization;
using System.Runtime.Serialization.Formatters.Binary;
using System.Text;

namespace ObycejnyGraf
{
  /// <summary>
  /// Binární datová vrstva komponenty <see cref="ObycejnyGraf"/>.
  /// Umožňuje načítat a ukládat části komponenty.
  /// </summary>
  sealed class BinarniDatovyModul : IDatovyModul
  {
    #region Datové složky
    private static BinarniDatovyModul binarniDatovyModul = new BinarniDatovyModul();
    #endregion

    #region Vlastnosi
    /// <summary>
    /// Zpřístupní jedináčka datového modulu.
    /// </summary>
    public static BinarniDatovyModul GetInstance { get { return binarniDatovyModul; } }
    #endregion

    #region Metody
    /// <summary>
    /// Uloží graf jako binární soubor.
    /// <exception cref="System.ArgumentException">Metoda vyvolá výjimku, pokud soubor nelze vytvořit,
    /// nebo pokud graf nelze serializovat.</exception>
    /// </summary>
    /// <param name="cestaNazev">Cesta a název souboru.</param>
    /// <param name="graf">Graf, který se má uložit.</param>
    public void UlozGraf(string cestaNazev, GGraf graf)
    {
      IFormatter formatter = new BinaryFormatter();
      Stream stream = null;

      try
      {
        stream = new FileStream(cestaNazev, FileMode.Create, FileAccess.Write, FileShare.None);
        formatter.Serialize(stream, graf);
      }
      catch (Exception) { throw; }
      finally
      {
        if (stream != null)
        {
          stream.Close();
        }

      }
    }

    /// <summary>
    /// Načte graf z binárního souboru.
    /// <exception cref="System.ArgumentException">Metoda vyvolá výjimku, pokud soubor neexistuje,
    /// nebo pokud soubor nelze deserializovat.</exception>
    /// </summary>
    /// <param name="cestaNazev">Cesta a název souboru.</param>
    /// <returns>Načtený graf.</returns>
    public GGraf NactiGraf(string cestaNazev)
    {
      IFormatter formatter = new BinaryFormatter();
      Stream stream = null;
      try
      {
        stream = new FileStream(cestaNazev, FileMode.Open, FileAccess.Read, FileShare.Read);
        GGraf g = formatter.Deserialize(stream) as GGraf;
        return g;
      }
      catch (Exception)
      {
        throw;
      }
      finally
      {
        if (stream != null)
          stream.Close();
      }
    }

    /// <summary>
    /// Uloží nastavení komponenty <see cref="ObycejnyGraf"/>.
    /// </summary>
    /// <param name="cestaNazev">Místo uložení.</param>
    /// <param name="nastaveni">Přepravka s  nastavením.</param>
    internal void UlozNastaveni(string cestaNazev, NastaveniPrepravka nastaveni)
    {
      IFormatter formatter = new BinaryFormatter();
      Stream stream = new FileStream(cestaNazev, FileMode.Create, FileAccess.Write, FileShare.None);
      formatter.Serialize(stream, nastaveni);
      stream.Close();
    }

    /// <summary>
    /// Načte nastavení komponenty <see cref="ObycejnyGraf"/> z binárního souboru.
    /// <exception cref="System.ArgumentException">Metoda vyvolá výjimku, pokud soubor neexistuje, nebo pokud soubor nelze deserializovat.</exception>
    /// </summary>
    /// <param name="cestaNazev">Místo s nastavením.</param>
    /// <returns>Načtené nastavení.</returns>
    internal NastaveniPrepravka NactiNastaveni(string cestaNazev)
    {
      IFormatter formatter = new BinaryFormatter();
      Stream stream = null;
      try
      {
        stream = new FileStream(cestaNazev, FileMode.Open, FileAccess.Read, FileShare.Read);
        return formatter.Deserialize(stream) as NastaveniPrepravka;
      }
      catch (Exception)
      {
        throw;
      }
      finally
      {
        if (stream != null)
          stream.Close();
      }
    }
    #endregion

    #region Konstruktor
    private BinarniDatovyModul() { }
    #endregion
  }
}
