﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using MG.DatoveStruktury.Grafy.SmisenaReprezentace;
using MG.DatoveStruktury.Tabulky;

namespace ObycejnyGraf
{
  /// <summary>
  /// Graf je neorientovaný, obyčejný => nepodporuje násobné hrany ani smyčky. 
  /// Iterátor provádí prohlídku vrcholů do hloubky.
  /// <remarks>Prioritní je práce s vrcholy grafu (průměrná složitost 
  /// vkládání/zpřístupňování/odstraňování log(n) o základu 2,
  /// kde n je počet vrcholů v grafu. Práce s hranami je druhotní.
  /// Graf je upraven pro grafickou prezentaci v 2D souřadném systému.</remarks>
  /// </summary>
  [Serializable]
  public class GGraf : AdsGrafSmisenaReprezentace<GVrchol, GHrana, double, double, double>
  {
    #region Datové složky
    [NonSerialized]
    private IVlastnikGrafu vlastnik;
    #endregion


    #region Vlastnosti
    /// <summary>
    /// Vlastník grafu. Vyplňuje se v případě, že je vlastníkem objekt typu <see cref="IPrvekGrafu"/>
    /// a využívá graf ke grafické prezentaci dat.
    /// </summary>
    public IVlastnikGrafu Vlastnik
    { get { return vlastnik; } set { vlastnik = value; } }
    #endregion


    #region Metody
    /// <summary>
    /// Vloží vrchol do struktury grafu.
    /// <exception cref="System.ArgumentException">Metoda je chráněna výjimkou pro případ, 
    /// že by se přidával duplicitní vrchol.</exception>
    /// </summary>
    /// <param name="vrchol">Objekt <see cref="GVrchol"/>, který se vloží do grafu.</param>
    public virtual void VlozVrchol(GVrchol vrchol)
    {
      try
      {
        base.VlozVrchol(vrchol, vrchol.Klic);
      }
      catch (Exception)
      {
        throw;
      }

      (vrchol as GVrchol).vlastnikVrcholu = this;

      if (this.Vlastnik != null)
      {
        Vlastnik.Invalidate();
      }
    }

    /// <summary>
    /// Vyhledá vrchol ve struktuře grafu podle klíče.
    /// </summary>
    /// <param name="klic">Klíč vrcholu.</param>
    /// <returns>Vrchol dle klíče nebo hodnota <c>null</c>.</returns>
    new public virtual GVrchol NajdiVrchol(double klic) { return (GVrchol)base.NajdiVrchol(klic); }

    /// <summary>
    /// Pokud má graf vlastníka, upraví graficky stávající vrchol.
    /// </summary>
    /// <param name="klic">Klíč vrcholu.</param>
    public virtual void UpravVrchol(double klic)
    {
      if (Vlastnik != null)
      {
        GVrchol vrchol = NajdiVrchol(klic);
        if (vrchol != null)
          Vlastnik.UpravVrchol(vrchol);
      }
    }

    /// <summary>
    /// Pokud má graf vlastníka, upraví graficky stávající hranu.
    /// </summary>
    /// <param name="klic">Klíč hrany.</param>
    public virtual void UpravHranu(double klic)
    {
      if (Vlastnik != null)
      {
        GHrana hrana = NajdiHranu(klic);
        if (hrana != null)
          Vlastnik.UpravHranu(hrana);
      }
    }

    /// <summary>
    /// Vyhledá vrchol dle klíče, následně se ho pokusí odstranit.
    /// <remarks>Metoda nekontroluje, zda stupeň vrcholu je 0.</remarks>
    /// </summary>    
    /// <param name="klic">Klíč odstraňovaného vrcholu.</param>
    /// <returns>Odstraněný vrchol nebo hodnota <c>null</c> v případě, 
    /// že daný vrchol neexistuje.</returns>
    new public virtual GVrchol OdstranVrchol(double klic)
    {
      Vrchol<double, double, double> v = base.OdstranVrchol(klic);
      if (Vlastnik != null)
        Vlastnik.Invalidate();
      return v as GVrchol;
    }

    /// <summary>
    /// Vloží hranu do grafu dle zadaných parametrů.
    /// </summary>
    /// <param name="hrana">Objekt typu <see cref="GHrana"/> s požadovanými atributy.</param>
    /// <param name="klic">Klíč hrany.</param>
    /// <exception cref="System.ArgumentException">Metoda vyvolá výjimku v případě, že se vkládá násobná hrana, 
    /// smyčka, nebo některý z incidenčních vrcholů není součástí množiny vrcholů grafu.</exception>
    public virtual void VlozHranu(GHrana hrana, double klic)
    {
      GVrchol iv1 = base.NajdiVrchol(hrana.IncidencniVrchol1.Klic) as GVrchol;
      GVrchol iv2 = base.NajdiVrchol(hrana.IncidencniVrchol2.Klic) as GVrchol;

      if (ReferenceEquals(hrana.IncidencniVrchol1, hrana.IncidencniVrchol2))
      {
        throw new ArgumentException("Graf nepodporuje smyčky.");
      }

      foreach (Hrana<double, double, double> item in hrana.IncidencniVrchol1.SeznamSousedu)
      {
        if (ReferenceEquals(hrana.IncidencniVrchol2, item.IncidencniVrchol1)||
          ReferenceEquals(hrana.IncidencniVrchol2, item.IncidencniVrchol2))
        {
          throw new ArgumentException("Graf nepodporuje násobné hrany.");
        }
      }

      try
      {
        base.VlozHranu(hrana.IncidencniVrchol1.Klic,
          hrana.IncidencniVrchol2.Klic, hrana, klic);
      }
      catch (Exception)
      {
        throw;
      }

      (hrana as GHrana).vlastnikHrany = this;

      if (this.Vlastnik != null)
      {
        Vlastnik.Invalidate();
      }
    }

    /// <summary>
    /// Vyhledá hranu pomocí daných klíčů.
    /// </summary>
    /// <param name="klic">Klíč hrany.</param>
    /// <returns>Požadovanou hranu, nebo hodnotu <c>null</c> v případě že daná hrana neexistuje.</returns>
    new public virtual GHrana NajdiHranu(double klic)
    {
      return (GHrana)base.NajdiHranu(klic);
    }

    /// <summary>
    /// Vyhledá hranu dle klíčů, a následně se pokusí hranu odstranit.
    /// </summary>
    /// <param name="klic">Klíč hrany.</param>
    /// <returns>Odstraněnou hranu, nebo hodnotu <c>null</c> v případě, 
    /// že hrana s danými inc. vrcholy neexistuje.</returns>
    new public virtual GHrana OdstranHranu(double klic)
    {
      Hrana<double, double, double> h = base.OdstranHranu(klic);
      if (Vlastnik != null)
        Vlastnik.Invalidate();
      return h as GHrana;
    }

    //public new AdsTabulkaSeznam<GHrana, double> SeznamHran
    //{
    //  get
    //  {
    //    return (AdsTabulkaSeznam<GHrana, double>)base.SeznamHran;
    //  }
    //}
    #endregion


    #region Konstruktor
    /// <summary>
    /// Bezparametrický konstruktor
    /// </summary>
    public GGraf() { }

    /// <summary>
    /// Parametrický konstruktor. Inicializuje datovou složku vlastník,
    /// která slouží pro překreslení mateřské komponenty při změně struktury grafu.
    /// </summary>
    /// <param name="vlastnik">Mateřská komponenta typu <see cref="System.Windows.Forms.Control"/> 
    /// nebo jeho potomka.</param>
    public GGraf(IVlastnikGrafu vlastnik) { this.Vlastnik = vlastnik; }
    #endregion
  }
}
