﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using MG.DatoveStruktury.Grafy.SmisenaReprezentace;

namespace ObycejnyGraf
{
  /// <summary>
  /// Obsahuje informace pro ukládání do sekundární datové struktury uchovávající neorientované grafy (vrcholový přístup).
  /// <remarks>Hrana je upravena pro grafickou prezentaci grafu.</remarks>
  /// </summary>
  [Serializable]
  public class GHrana : Hrana<double, double, double>
  {
    #region Datové složky
    private static double k = -99999999999999;

    private double klic;

    private string nazev;

    /// <summary>
    /// Vlastník hrany. Při změně klíče hrany je nutné danou hranu odebrat
    /// ze struktury grafu a znovu vložit.
    /// </summary>
    internal GGraf vlastnikHrany = null;

    private System.Drawing.Color barva;

    private int tloustka;

    private System.Drawing.Font fontTextu;

    private System.Drawing.Color barvaTextu;

    private bool zobrazitText;

    /// <summary>
    /// Informace, zda je hrana aktuální hranou.
    /// </summary>
    internal bool JeAktivni;
    #endregion


    #region Vlastnosti
    /// <summary>
    /// Příští hodnota, která se přiřadí jako implicitní klíč hrany. Hodnota klíče postupně stoupá o 1.
    /// </summary>
    public static double ImplicitniKlic { get { return k; } internal set { k = value; } }
    /// <summary>
    /// Klíč hrany. Při změně klíče hrany je nutné danou hranu odebrat
    /// ze struktury grafu a znovu vložit (složitost 2 * (2 * (log(n) o základu 2 + O(s)) + O(m)),
    /// kde n je počet vrcholů v grafu, m je počet hranu v grafu, a s je počet hran incidujících s vrcholem,
    /// se kterým inciduje daná hrana).
    /// </summary>
    public double Klic
    {
      get { return klic; }
      set
      {
        if (value != klic)
        {
          GHrana gh = vlastnikHrany.OdstranHranu(Klic);
          klic = value;
          vlastnikHrany.VlozHranu(this, Klic);
        }
      }
    }

    /// <summary>
    /// Název hrany. Implicitně (název inc. vrcholu 1, název inc. vrcholu 2).
    /// </summary>
    public string Nazev
    {
      get { return this.nazev; }
      set 
      {
        if (nazev != value)
        {
          this.nazev = value;
          if (vlastnikHrany != null && vlastnikHrany.Vlastnik != null)
            vlastnikHrany.Vlastnik.Invalidate();
        }
      }
    }

    /// <summary>
    /// Ohodnocení hrany.
    /// </summary>
    public override double Ohodnoceni 
    { 
      get { return base.Ohodnoceni; }
      set { if (base.Ohodnoceni != value) { base.Ohodnoceni = value; if (vlastnikHrany != null && vlastnikHrany.Vlastnik != null) 
        vlastnikHrany.Vlastnik.Invalidate(); } }
    }

    /// <summary>
    /// První incidenční vrchol hrany.
    /// </summary>
    new public GVrchol IncidencniVrchol1
    {
      get { return (GVrchol)base.IncidencniVrchol1; }
      //private set;
    }

    /// <summary>
    /// Druhý incidenční vrchol hrany.
    /// </summary>
    new public GVrchol IncidencniVrchol2
    {
      get { return (GVrchol)base.IncidencniVrchol2; }
      //private set;
    }

    /// <summary>
    /// Barva hrany při vykreslení na kreslící plátno komponenty.
    /// </summary>
    public System.Drawing.Color Barva
    {
      get { return this.barva; }
      set 
      {
        if (barva != value)
        {
          this.barva = value;
          if (vlastnikHrany != null && vlastnikHrany.Vlastnik != null) vlastnikHrany.Vlastnik.Invalidate();
        }
      }
    }

    /// <summary>
    /// Tloušťka hrany v pixelech při vykreslování.
    /// </summary>
    public int Tloustka
    {
      get { return this.tloustka; }
      set { if (tloustka != value) { this.tloustka = value; if (vlastnikHrany != null && vlastnikHrany.Vlastnik != null) 
        vlastnikHrany.Vlastnik.Invalidate(); } }
    }

    /// <summary>
    /// Font ohodnocení hrany při zobrazování v komponentě.
    /// </summary>
    public System.Drawing.Font FontTextu
    {
      get { return fontTextu; }
      set { if (fontTextu != value) { fontTextu = value; if (vlastnikHrany != null && vlastnikHrany.Vlastnik != null) 
        vlastnikHrany.Vlastnik.Invalidate(); } }
    }

    /// <summary>
    /// Barva textu. 
    /// </summary>
    public System.Drawing.Color BarvaTextu
    {
      get { return barvaTextu; }
      set { if (barvaTextu != value) { barvaTextu = value; if (vlastnikHrany != null && vlastnikHrany.Vlastnik != null) 
        vlastnikHrany.Vlastnik.Invalidate(); } }
    }

    /// <summary>
    /// Hodnota určující, zda se má ohodnocení hrany zobrazit, či nikoliv.
    /// </summary>
    public bool ZobrazitText
    {
      get { return zobrazitText; }
      set { if (zobrazitText != value) { zobrazitText = value; if (vlastnikHrany != null && vlastnikHrany.Vlastnik != null) 
        vlastnikHrany.Vlastnik.Invalidate(); } }
    }
    #endregion

    #region Metody
    /// <summary>
    /// Řetězcová reprezentace prvku.
    /// </summary>
    /// <returns>Řetězcovou prezentaci prvku.</returns>
    public override string ToString()
    {
      return Nazev;
    }
    #endregion


    #region Konstruktor
    /// <summary>
    /// Parametrický konstruktor. Inicializuje datové složky hrany.
    /// </summary>
    /// <param name="incidencniVrchol1">První incidenční vrchol.</param>
    /// <param name="incidencniVrchol2">Druhý incidenční vrchol.</param>
    /// <param name="klic">Klíč hrany. Implicitně se přidělí automatický klíč.</param>
    public GHrana(GVrchol incidencniVrchol1, GVrchol incidencniVrchol2, double klic = double.MaxValue)
      : base(incidencniVrchol1, incidencniVrchol2) 
    {
      if (klic == double.MaxValue)
        this.klic = k++;
      else
        this.klic = klic;

      Nazev = "(" + incidencniVrchol1.Nazev + ", " + incidencniVrchol2.Nazev + ")";
    }
    #endregion
  }
}
