﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using MG.DatoveStruktury.Grafy.SmisenaReprezentace;

namespace ObycejnyGraf
{
  /// <summary>
  /// Obsahuje informace pro ukládání do primární datové struktury uchovávající neorientované grafy (vrcholový přístup).
  /// <remarks>Objekt je upravený pro grafickou reprezentaci v 2D souřadném systému.
  /// Klíčem vrcholu jsou jeho souřadnice, což je měnitelná položka. To znamená, že při změně souřadnic
  /// se vrchol musí z grafu odstranit a vložit s novým klíčem (složitost průměrně 2 * log(n) o základu 2),
  /// kde n je počet vrcholů v grafu.
  /// Důvodem volby souřadnic jako klíče byla skutečnost, že graf se využívá k interaktivnímu vykreslování 
  /// v 2D souřadném systému, a vrcholy grafu se tak můžou vyhledávat na základě pozice kurzoru v souřadném systému.</remarks>
  /// </summary>
  [Serializable]
  public class GVrchol : Vrchol<double, double, double>
  {
    #region Datové složky
    /// <summary>
    /// Statická proměnná pro přiřazení implicitního názvu vrcholu.
    /// </summary>
    private static int i = 0;

    private string nazev = "v" + i++.ToString();

    private double klic;

    /// <summary>
    /// Vlastník vrcholu. Při změně souřadnic vrcholu je nutné daný vrchol odebrat
    /// z primární struktury grafu a znovu vložit, jelikož klíč vrcholu je dán jeho souřadnicemi
    /// (složitost průměrně 2 * log(n)).
    /// </summary>
    internal GGraf vlastnikVrcholu = null;

    private SouradniceVrcholu souradnice;

    private System.Drawing.Font fontTextu;

    private System.Drawing.Color barvaTextu;

    private bool zobrazitText;

    private int velikost;

    private System.Drawing.Color barvaOhraniceni;

    private System.Drawing.Color barvaVyplne;

    /// <summary>
    /// Informace, zda je vrchol aktivním vrcholem.
    /// </summary>
    internal bool JeAktivni;
    #endregion


    #region Vlastnosti
    /// <summary>
    /// Název vrcholu. Implicitně v + pořadové číslo vrcholu.
    /// </summary>
    public string Nazev
    {
      get { return nazev; }
      set
      {
        if (nazev != value)
        {
          nazev = value;
          if (vlastnikVrcholu != null && vlastnikVrcholu.Vlastnik != null)
            vlastnikVrcholu.Vlastnik.Invalidate();
        }
      }
    }

    /// <summary>
    /// Souřadnice vrcholu.
    /// Při změně souřadnic vrcholu je nutné daný vrchol odebrat
    /// z primární struktury grafu a znovu vložit, jelikož klíč vrcholu je dán jeho souřadnicemi
    /// (složitost průměrně 2 * log(n), kde n je počet vrcholů v grafu).
    /// </summary>
    public SouradniceVrcholu Souradnice
    {
      get { return souradnice; }
      set
      {
        if (souradnice.X != value.X || souradnice.Y != value.Y)
        {
          GVrchol gv = this;
          if (vlastnikVrcholu != null)
          {
            vlastnikVrcholu.OdstranVrchol(Klic);
            souradnice = value;
            klic = GetKlic(Souradnice.X, Souradnice.Y);
            vlastnikVrcholu.VlozVrchol(gv);
          }
          else
          {
            souradnice = value;
            klic = GetKlic(Souradnice.X, Souradnice.Y);
            souradnice = value;
          }
          if (vlastnikVrcholu != null && vlastnikVrcholu.Vlastnik != null)
            vlastnikVrcholu.Vlastnik.Invalidate();
        }
      }
    }

    /// <summary>
    /// Klíč vrcholu.
    /// </summary>
    public double Klic
    {
      get { return klic; }
    }

    /// <summary>
    /// Font názvu vrcholu při zobrazování v komponentě.
    /// </summary>
    public System.Drawing.Font FontTextu
    {
      get { return fontTextu; }
      set { if (fontTextu != value) { fontTextu = value; if (vlastnikVrcholu != null && vlastnikVrcholu.Vlastnik != null) 
        vlastnikVrcholu.Vlastnik.Invalidate(); } }
    }

    /// <summary>
    /// Barva textu.
    /// </summary>
    public System.Drawing.Color BarvaTextu
    {
      get { return barvaTextu; }
      set { if (barvaTextu != value) { barvaTextu = value; if (vlastnikVrcholu != null && vlastnikVrcholu.Vlastnik != null) 
        vlastnikVrcholu.Vlastnik.Invalidate(); } }
    }

    /// <summary>
    /// Hodnota určující, zda se má název vrcholu zobrazit, či nikoliv.
    /// </summary>
    public bool ZobrazitText
    {
      get { return zobrazitText; }
      set { if (zobrazitText != value) { zobrazitText = value; if (vlastnikVrcholu != null && vlastnikVrcholu.Vlastnik != null) 
        vlastnikVrcholu.Vlastnik.Invalidate(); } }
    }

    /// <summary>
    /// Velikost vrcholu v pixelech.
    /// </summary>
    public int Velikost
    {
      get { return velikost; }
      set { if (velikost != value) { velikost = value; if (vlastnikVrcholu != null && vlastnikVrcholu.Vlastnik != null) 
        vlastnikVrcholu.Vlastnik.Invalidate(); } }
    }

    /// <summary>
    /// Barva ohraničení vrcholu.
    /// </summary>
    public System.Drawing.Color BarvaOhraniceni
    {
      get { return barvaOhraniceni; }
      set { if (barvaOhraniceni != value) { barvaOhraniceni = value; if (vlastnikVrcholu != null && vlastnikVrcholu.Vlastnik != null) 
        vlastnikVrcholu.Vlastnik.Invalidate(); } }
    }

    /// <summary>
    /// Barva výplně vrcholu.
    /// </summary>
    public System.Drawing.Color BarvaVyplne
    {
      get { return barvaVyplne; }
      set { if (barvaVyplne != value) { barvaVyplne = value; if (vlastnikVrcholu != null && vlastnikVrcholu.Vlastnik != null) 
        vlastnikVrcholu.Vlastnik.Invalidate(); } }
    }
    #endregion


    #region Metody
    /// <summary>
    /// Vypočte jedinečný klíč vrcholu z jeho souřadnic.
    /// </summary>
    /// <param name="sX">X-ová souřadnice vrcholu.</param>
    /// <param name="sY">Y-ová souřadnice vrcholu.</param>
    /// <returns>Vypočtený klíč vrcholu.</returns>
    protected virtual double GetKlic(double sX, double sY)
    {
      int posun = 1;
      int x = (int)sX;
      int y = (int)sY;
      double klic = Math.Abs(y);

      int _y = Math.Abs(y);
      while ((_y % 10) > 0)
      {
        posun *= 10;
        _y /= 10;
      }

      klic = posun * ((y < 0) ? 3 : 4) + klic;
      posun *= 10;
      klic = posun * Math.Abs(x) + klic;

      int _x = Math.Abs(x);
      while ((_x % 10) > 0)
      {
        posun *= 10;
        _x /= 10;
      }

      klic = posun * ((x < 0) ? 1 : 2) + klic;

      return klic;
    }

    /// <summary>
    /// Řetězcová reprezentace prvku.
    /// </summary>
    /// <returns>Řetězcovou prezentaci prvku.</returns>
    public override string ToString()
    {
      return Nazev;
    }
    #endregion


    #region Konstruktor
    /// <summary>
    /// Parametrický konstruktor. Inicializuje datové složky vrcholu.
    /// </summary>
    /// <param name="souradnice">Souřadnice vrcholu.</param>
    public GVrchol(SouradniceVrcholu souradnice)
    {
      this.souradnice = souradnice;
      klic = GetKlic(Souradnice.X, Souradnice.Y);
    }
    #endregion
  }
}
