﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using MG.Matematika.Geometrie2D;
using KresliciPlatno;
using MG.DatoveStruktury.Grafy.SmisenaReprezentace;
using System.Runtime.Serialization;
using System.Runtime.Serialization.Formatters.Binary;
using System.IO;
using System.Diagnostics;
using MG.DatoveStruktury.Tabulky;


namespace ObycejnyGraf
{
  /// <summary>
  /// Komponenta vytvořena za účelem zobrazování, poskytování a úchovu obyčejného grafu. 
  /// Je možnost povolit grafickou editaci struktury grafu
  /// skrz atribut <see cref="GrafickaEditaceStruktury"/>.
  /// Dále je možné povolit vykreslovvání aktivního prvku grafu pomocí vlastnosti 
  /// <see cref="VykreslovaniAktivnihoPrvku"/>. Pomocí vlastnosti <see cref="CestaKeKonfiguracnimuSouboru"/> lze 
  /// načítat a ukládat konfiguraci komponenty, pokud je nastavena vlastnost 
  /// <see cref="AutomatickeUkladaniNacitaniKonfigurace"/>.
  /// </summary>
  public partial class ObycejnyGraf : UserControl, IVlastnikGrafu
  {
    #region Datové složky
    private bool grafickaEditaceStruktury = true;
    private bool vykreslovaniAktivnihoPrvku = true;
    private string cestaKeKonfiguracnimuSouboru = "nastaveni.bin";

    private Keys controlPridavaciRezim = Keys.Control;
    private Keys controlOdstranovaciRezim = Keys.Alt;
    private Keys controlEditacniRezim = Keys.Shift;

    private SouradnySystem2D ss;
    private GGraf graf;
    private Bod<int> souradniceKurzoru;

    /// <summary>
    /// Hodnota určující režim posouvání ss.
    /// </summary>
    private bool rezRoll;
    /// <summary>
    /// Pomocná proměnná pro zjištění směru a míry posunu kurzoru v režimu rezRoll.
    /// </summary>
    private Bod<int> rollovaciSouradnice = new Bod<int>();
    /// <summary>
    /// Hodnota určující režim přidávání hrany.
    /// </summary>
    private bool rezPridHra;
    /// <summary>
    /// Pomocný atribut pro grafické přidání hrany do grafu.
    /// První incidenční vrchol hrany, pomocná proměnná režimu rezPridHra.
    /// </summary>
    private GVrchol iv1;
    /// <summary>
    /// Pomocný atribut pro grafické přidání hrany do grafu.
    /// Druhý incidenční vrchol hrany, pomocná proměnná režimu rezPridHra.
    /// </summary>
    private GVrchol iv2;
    /// <summary>
    /// Typ editace komponenty. Editace vrcholů nebo editace hran.
    /// </summary>
    private TypRezmimuPrace rezimPrace = TypRezmimuPrace.PraceSVrcholy;
    /// <summary>
    /// Aktuální prvek komponenty.
    /// </summary>
    private IPrvekGrafu aktualniPrvek;
    /// <summary>
    /// Preaktuální prvek komponenty.
    /// </summary>
    private IPrvekGrafu preAktualniPrvek;
    #endregion


    #region Události
    /// <summary>
    /// Událost je spuštěna při změně struktury grafu (odstranění/vložení/úprava vrcholu/hrany).
    /// </summary>
    [Category("Události")]
    [Description("Událost je spuštěna při změně struktury grafu (odstranění/vložení/úprava vrcholu/hrany)."),
    Browsable(true)]
    public event ZmenaStrukturyGrafuEventHandler ZmenaStrukturyGrafu;

    /// <summary>
    /// Událost je spuštěna před pokusem o odstranění vrcholu/hrany).
    /// </summary>
    [Category("Události")]
    [Description("Událost je spuštěna před pokusem o odstranění vrcholu/hrany."),
    Browsable(true)]
    public event ZmenaStrukturyGrafuCancelEventHandler OdstraneniPrvku;
    #endregion


    #region Vlastnosti
    /// <summary>
    /// <see cref="bool"/> Umožní grafickou editaci struktury grafu (Komponenta komunikuje s okolím pomocí událostí).
    /// </summary>
    [Description("Umožní grafickou editaci struktury grafu (Komponenta komunikuje s okolím pomocí událostí)."),
    Category("Konfigurace komponenty"),
    DefaultValue(true),
    Browsable(true)]
    public bool GrafickaEditaceStruktury
    {
      get { return grafickaEditaceStruktury; }
      set { grafickaEditaceStruktury = value; }
    }

    /// <summary>
    /// <see cref="bool"/> Umožní grafické vykreslování aktivního prvku grafu dle platného režimu práce.
    /// </summary>
    [Description("Umožní grafické vykreslování aktivního prvku grafu dle platného režimu práce."),
    Category("Konfigurace komponenty"),
    DefaultValue(true),
    Browsable(true)]
    public bool VykreslovaniAktivnihoPrvku
    {
      get { return vykreslovaniAktivnihoPrvku; }
      set { vykreslovaniAktivnihoPrvku = value; }
    }

    /// <summary>
    /// <see cref="bool"/> Umožní automatické ukládání konfiguračních údajů při jejich změně do konfiguračního souboru
    /// dle vlastnosti <see cref="CestaKeKonfiguracnimuSouboru"/>. Dále umožňuje automatické načítání z tohoto souboru 
    /// při startu aplikace
    /// </summary>
    [Description("Umožní automatické ukládání konfiguračních údajů při jejich změně do konfiguračního souboru" +
      "dle vlastnosti CestaKeKonfiguracnimuSouboru. Dále umožňuje automatické načítání z tohoto souboru" +
      "při startu aplikace"),
    Category("Konfigurace komponenty"),
    DefaultValue(true),
    Browsable(true)]
    public bool AutomatickeUkladaniNacitaniKonfigurace { get; set; }

    /// <summary>
    ///Cesta ke konfiguračnímu souboru komponenty s koncovkou *.bin.
    ///V případě, že soubor neexistuje, použije se implicitní konfigurace komponenty.
    /// </summary>
    [Description("Cesta ke konfiguračnímu souboru komponenty s koncovkou *.bin." +
      "V případě, že soubor neexistuje, použije se implicitní konfigurace komponenty."),
    Category("Konfigurace komponenty"),
    DefaultValue("nastaveni.bin"),
    Browsable(true)]
    public string CestaKeKonfiguracnimuSouboru
    {
      get { return cestaKeKonfiguracnimuSouboru; }
      set
      {
        if (value.Contains(".bin"))
          cestaKeKonfiguracnimuSouboru = value;
        else
          return;

        if (AutomatickeUkladaniNacitaniKonfigurace)
          try
          {
            RozbalPrepravku(BinarniDatovyModul.
              GetInstance.NactiNastaveni(CestaKeKonfiguracnimuSouboru));
            VykresliPictureBoxy();
            kpPlatno.Invalidate();
          }
          catch (Exception)
          {
            MessageBox.Show("Soubor "+value+" nebyl doposud vytvořen nebo není validní konfigurací komponenty. Použije se" +
              "Implicitní nastavení. Po změně konfigurace se nová konfigurace automaticky uloží.",
              "Automatické načtení konfiguračního souboru", MessageBoxButtons.OK, MessageBoxIcon.Information);
          }
      }
    }

    /// <summary>
    /// Barva osy komponenty.
    /// </summary>
    [Description("Barva osy komponenty."),
    Category("Konfigurace komponenty"),
    DefaultValue(typeof(Color), "Gray"),
    Browsable(true)]
    public Color BarvaOsy { get; set; }

    /// <summary>
    /// Vnější barva vrcholu.
    /// </summary>
    [Description("Vnější barva vrcholu."),
    Category("Konfigurace komponenty"),
    DefaultValue(typeof(Color), "Blue"),
    Browsable(true)]
    public Color VnejsiBarvaVrcholu { get; set; }

    /// <summary>
    /// Vnitřní barva vrcholu.
    /// </summary>
    [Description("Vnitřní barva vrcholu."),
    Category("Konfigurace komponenty"),
    DefaultValue(typeof(Color), "White"),
    Browsable(true)]
    public Color VnitrniBarvaVrcholu { get; set; }

    /// <summary>
    /// Barva hrany.
    /// </summary>
    [Description("Barva hrany."),
    Category("Konfigurace komponenty"),
    DefaultValue(typeof(Color), "Black"),
    Browsable(true)]
    public Color BarvaHrany { get; set; }

    /// <summary>
    /// Barva textu vrcholu.
    /// </summary>
    [Description("Barva textu vrcholu."),
    Category("Konfigurace komponenty"),
    DefaultValue(typeof(Color), "Black"),
    Browsable(true)]
    public Color BarvaTextuVrcholu { get; set; }

    /// <summary>
    /// Barva textu hrany.
    /// </summary>
    [Description("Barva textu hrany."),
    Category("Konfigurace komponenty"),
    DefaultValue(typeof(Color), "Black"),
    Browsable(true)]
    public Color BarvaTextuHrany { get; set; }

    /// <summary>
    /// Barva aktivního prvku.
    /// </summary>
    [Description("Barva aktivního prvku."),
    Category("Konfigurace komponenty"),
    DefaultValue(typeof(Color), "Red"),
    Browsable(true)]
    public Color BarvaAktivnihoPrvku { get; set; }

    /// <summary>
    /// Velikost vrcholu v pixelech.
    /// </summary>
    [Description("Velikost vrcholu v pixelech."),
    Category("Konfigurace komponenty"),
    DefaultValue(10),
    Browsable(true)]
    public int VelikostVrcholu { get; set; }

    /// <summary>
    /// Font vrcholu.
    /// </summary>
    [Description("Font vrcholu."),
    Category("Konfigurace komponenty"),
      //DefaultValue(typeof(Font), new Font("arial", 9).ToString()),
    Browsable(true)]
    public Font FontVrcholu { get; set; }

    /// <summary>
    /// Zobrazení (vykreslení) textu vrcholu.
    /// </summary>
    [Description("Zobrazení (vykreslení) textu vrcholu."),
    Category("Konfigurace komponenty"),
    DefaultValue(true),
    Browsable(true)]
    public bool ZobrazitTextVrcholu { get; set; }

    /// <summary>
    /// Tloušťka hrany v pixelech.
    /// </summary>
    [Description("Tloušťka hrany v pixelech."),
    Category("Konfigurace komponenty"),
    DefaultValue(3),
    Browsable(true)]
    public int TloustkaHrany { get; set; }

    /// <summary>
    /// Ohodnocení hrany.
    /// </summary>
    [Description("Ohodnocení hrany."),
    Category("Konfigurace komponenty"),
    DefaultValue(1),
    Browsable(true)]
    public double OhodnoceniHrany { get; set; }

    /// <summary>
    /// Font hrany.
    /// </summary>
    [Description("Font hrany."),
    Category("Konfigurace komponenty"),
      //DefaultValue(typeof(Font), new Font("arial", 9).ToString()),
    Browsable(true)]
    public Font FontHrany { get; set; }

    /// <summary>
    /// Zobrazení (vykreslení) textu hrany.
    /// </summary>
    [Description("Zobrazení (vykreslení) textu hrany."),
    Category("Konfigurace komponenty"),
    DefaultValue(true),
    Browsable(true)]
    public bool ZobrazitTextHrany { get; set; }

    /// <summary>
    /// Režim práce komponenty.
    /// </summary>
    [Browsable(false)]
    [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
    public TypRezmimuPrace RezimPrace
    {
      get { return rezimPrace; }
      set
      {
        if (rezimPrace != value)
          pbTypEditace_Click(kpPlatno, null);
      }
    }

    /// <summary>
    /// Graf, který se vykresluje, edituje a uchovává.
    /// </summary>
    [Browsable(false)]
    [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
    public GGraf Graf
    {
      get { return graf; }

      private set
      {
        this.graf = value;
        graf.Vlastnik = this;
        kpPlatno.Invalidate();
        GHrana.ImplicitniKlic += value.SeznamHran.Mohutnost + 1;
      }
    }

    /// <summary>
    /// Systém vykreslování objektů na kreslící plátno. Obsahuje základní prostředky pro přepočty 
    /// souřadnic mezi reálným SS a pixelovým SS.
    /// </summary>
    [Browsable(false)]
    [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
    public SouradnySystem2D SouradnySystem
    {
      get { return ss; }
    }

    /// <summary>
    /// Aktuální souřadnice kurzoru (reálné souřadnice, zaokrouhlené na celé číslo).
    /// </summary>
    [Browsable(false)]
    [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
    public Bod<int> SouradniceKurzoru
    {
      get { return souradniceKurzoru; }
    }

    /// <summary>
    /// Aktivní prvek grafu dle platného režimu práce, nad nímž se nachází kurzor.
    /// Musí být aktivovaná vlastnost <see cref="VykreslovaniAktivnihoPrvku"/>.
    /// </summary>
    [Browsable(false)]
    [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
    public IPrvekGrafu AktivniPrvek { get { return aktualniPrvek; } }
    #endregion


    #region Metody
    /// <summary>
    /// Krycí metoda pro event ZmenaStrukturyGarfu.
    /// </summary>
    /// <param name="sender">Objekt, který event vyvolal.</param>
    /// <param name="e">Doplňující informace.</param>
    protected void OnZmenaStrukturyGrafu(object sender, ZmenaStrukturyGrafuEventArgs e)
    {
      ZmenaStrukturyGrafuEventHandler handler = ZmenaStrukturyGrafu;
      if (handler != null)
      {
        handler(sender, e);
      }
    }

    /// <summary>
    /// Krycí metoda pro event Cancel.
    /// </summary>
    /// <param name="sender">Objekt, který event vyvolal.</param>
    /// <param name="e">Doplňující informace.</param>
    protected void OnOdstraneniPrvku(object sender, ZmenaStrukturyGrafuCancelEventArgs e)
    {
      ZmenaStrukturyGrafuCancelEventHandler handler = OdstraneniPrvku;
      if (handler != null)
      {
        handler(sender, e);
      }
    }

    /// <summary>
    /// Vykreslí vrchol s popiskem dle atributů.
    /// </summary>
    /// <param name="v"><see cref="GVrchol"/>, který se má vykreslit.</param>
    /// <param name="g">Objekt <see cref="Graphics"/>,</param>
    /// <param name="aktivni">Pokud se nastaví na <c>True</c>, překreslí se pouze barva vrcholu na aktivní/původní barvu.</param>
    private void VykresliVrchol(GVrchol v, Graphics g, bool aktivni = false)
    {
      Color b = (v.JeAktivni ? BarvaAktivnihoPrvku : v.BarvaVyplne);
      g.DrawEllipse(new Pen(v.BarvaOhraniceni, 1), ss.R2PX(v.Souradnice.X) - (v.Velikost / 2), ss.R2PY(v.Souradnice.Y) - (v.Velikost / 2),
        v.Velikost, v.Velikost);
      g.FillEllipse(new SolidBrush(b), ss.R2PX(v.Souradnice.X) - (v.Velikost / 2), ss.R2PY(v.Souradnice.Y) - (v.Velikost / 2),
         v.Velikost, v.Velikost);

      if (!aktivni)
      {
        if (v.ZobrazitText)
          g.DrawString(v.Nazev, v.FontTextu, new SolidBrush(v.BarvaTextu), new Point(ss.R2PX(v.Souradnice.X),
            ss.R2PY(v.Souradnice.Y) + (int)v.FontTextu.Size / 4 + (int)v.Velikost / 2));
      }
    }

    /// <summary>
    /// Vykreslí hranu s popiskem dle atributů.
    /// </summary>
    /// <param name="h"><see cref="GHrana"/>, která se má vykreslit.</param>
    /// <param name="g">Objekt <see cref="Graphics"/></param>
    /// <param name="aktivni">Pokud se nastaví na <c>True</c>, překreslí se pouze barva hrany na aktivní/původní barvu.</param>
    private void VykresliHranu(GHrana h, Graphics g, bool aktivni = false)
    {
      g.DrawLine(new Pen((h.JeAktivni ? BarvaAktivnihoPrvku : h.Barva), h.Tloustka),
        ss.R2PX(h.IncidencniVrchol1.Souradnice.X), ss.R2PY(h.IncidencniVrchol1.Souradnice.Y),
        ss.R2PX(h.IncidencniVrchol2.Souradnice.X), ss.R2PY(h.IncidencniVrchol2.Souradnice.Y));

      if (!aktivni)
      {
        if (h.ZobrazitText)
        {
          Bod<double> stred = FunkceGeometricke.StredUsecky(new Bod<double>(ss.R2PX(h.IncidencniVrchol1.Souradnice.X),
            ss.R2PY(h.IncidencniVrchol1.Souradnice.Y)),
          new Bod<double>(ss.R2PX(h.IncidencniVrchol2.Souradnice.X), ss.R2PY(h.IncidencniVrchol2.Souradnice.Y)));
          g.DrawString(h.Ohodnoceni.ToString(), h.FontTextu, new SolidBrush(h.BarvaTextu),
            (int)stred.X + 3, (int)stred.Y + 2);
        }
      }
      else
      {
        VykresliVrchol(h.IncidencniVrchol1, g, true);
        VykresliVrchol(h.IncidencniVrchol2, g, true);
      }
    }

    /// <summary>
    /// Vykreslí osu reálného ss.
    /// </summary>
    /// <param name="g">Objekt <see cref="Graphics"/></param>
    private void VykresliOsu(Graphics g)
    {
      Pen pG = new Pen(BarvaOsy, 1);
      g.DrawLine(pG, ss.R2PX(0), 0, ss.R2PX(0), kpPlatno.Height);
      g.DrawLine(pG, 0, ss.R2PY(0), kpPlatno.Width, ss.R2PY(0));
    }

    /// <summary>
    /// Vykreslí graf do ss.
    /// </summary>
    /// <param name="g">Objekt <see cref="Graphics"/></param>
    private void VykresliGraf(Graphics g)
    {
      foreach (IPrvekTabulky<GHrana, double> item in graf.SeznamHran)
      {
        VykresliHranu(item.Data as GHrana, g);
      }

      foreach (GVrchol vrchol in graf)
      {
        VykresliVrchol(vrchol, g);
      }
    }

    /// <summary>
    /// Vykreslí pozadí obrázkovým tlačítkům dle aktuálního nastavení.
    /// </summary>
    private void VykresliPictureBoxy()
    {
      pbNastaveni.BackColor = this.BackColor;
      pbInformace.BackColor = this.BackColor;
      pbTypEditace.BackColor = this.BackColor;
    }

    /// <summary>
    /// Nalezne aktuální prvek komponenty.
    /// </summary>
    /// <param name="souradniceKurzoru">Aktuální souřadnice kurzoru.</param>
    private void NajdiAktualniPrvek(Bod<double> souradniceKurzoru)
    {
      double nejmensiVzdalenost = double.MaxValue, vzdalenost;
      aktualniPrvek = null;
      if (rezimPrace == TypRezmimuPrace.PraceSVrcholy)
      {
        foreach (GVrchol item in Graf)
        {
          vzdalenost = FunkceGeometricke.DelkaUsecky(souradniceKurzoru, item.Souradnice);
          if (vzdalenost < nejmensiVzdalenost)
          {
            nejmensiVzdalenost = vzdalenost;
            aktualniPrvek = item;
          }
        }
      }
      else
      {
        foreach (IPrvekTabulky<GHrana, double> item in Graf.SeznamHran)
        {
          GHrana hrana = item.Data;
          vzdalenost = FunkceGeometricke.DelkaUsecky(souradniceKurzoru,
            FunkceGeometricke.StredUsecky(hrana.IncidencniVrchol1.Souradnice, hrana.IncidencniVrchol2.Souradnice));
          if (vzdalenost < nejmensiVzdalenost)
          {
            nejmensiVzdalenost = vzdalenost;
            aktualniPrvek = hrana;
          }
        }
      }
    }

    /// <summary>
    /// Změní vlastnosti hrany dle dialogového okna.
    /// </summary>
    /// <param name="h">Měněná hrana.</param>
    public void UpravHranu(GHrana h)
    {
      using (FrmHrana frm = new FrmHrana(h))
      {
        if (frm.ShowDialog() == System.Windows.Forms.DialogResult.OK)
        {
          h = frm.Hrana;
        }
      }
    }

    /// <summary>
    /// Změní vlastnosti vrcholu dle dialogového okna.
    /// </summary>
    /// <param name="v">Měněný vrchol.</param>
    public void UpravVrchol(GVrchol v)
    {
      using (FrmVrchol frm = new FrmVrchol(v))
      {
        if (frm.ShowDialog() == System.Windows.Forms.DialogResult.OK)
        {
          v = frm.Vrchol;
        }
      }
    }

    /// <summary>
    /// Změní konfigurační nastavení komponenty.
    /// </summary>
    /// <param name="nastaveni">Nová přepravka s nastavením.</param>
    private void ZmenNastaveni(NastaveniPrepravka nastaveni)
    {
      RozbalPrepravku(nastaveni);
      if (AutomatickeUkladaniNacitaniKonfigurace)
      {
        if (!cestaKeKonfiguracnimuSouboru.Contains(".bin"))
          throw new ArgumentException("Cesta ke konfiguračnímu souboru není validní (soubor s koncovkou *.bin).");

          BinarniDatovyModul.GetInstance.UlozNastaveni
            (CestaKeKonfiguracnimuSouboru, nastaveni);
      }
      VykresliPictureBoxy();
      kpPlatno.Invalidate();
    }

    /// <summary>
    /// Naplní přepravku s nastavením dle současné konfigurace komponenty.
    /// </summary>
    private NastaveniPrepravka NaplnPrepravku()
    {
      return new NastaveniPrepravka(BackColor, BarvaOsy, VnejsiBarvaVrcholu, VnitrniBarvaVrcholu,
        BarvaHrany, BarvaTextuVrcholu, BarvaTextuHrany, BarvaAktivnihoPrvku, VelikostVrcholu, FontVrcholu,
        ZobrazitTextVrcholu, TloustkaHrany, OhodnoceniHrany, FontHrany, ZobrazitTextHrany);
    }

    /// <summary>
    /// Nakonfiguruje komponentu dle předané konfigurační přepravky.
    /// </summary>
    /// <param name="nastaveni">Konfigurační přepravka.</param>
    private void RozbalPrepravku(NastaveniPrepravka nastaveni)
    {
      BackColor = nastaveni.BarvaPozadi;
      BarvaOsy = nastaveni.BarvaOsy;
      VnejsiBarvaVrcholu = nastaveni.VnejsiBarvaVrcholu;
      VnitrniBarvaVrcholu = nastaveni.VnitrniBarvaVrcholu;
      BarvaHrany = nastaveni.BarvaHrany;
      BarvaAktivnihoPrvku = nastaveni.BarvaAktivnihoPrvku;
      BarvaTextuVrcholu = nastaveni.BarvaTextuVrcholu;
      BarvaTextuHrany = nastaveni.BarvaTextuHrany;
      FontVrcholu = nastaveni.FontVrcholu;
      FontHrany = nastaveni.FontHrany;
      ZobrazitTextVrcholu = nastaveni.ZobrazitTextVrcholu;
      ZobrazitTextHrany = nastaveni.ZobrazitTextHrany;
      VelikostVrcholu = nastaveni.VelikostVrcholu;
      TloustkaHrany = nastaveni.TloustkaHrany;
      OhodnoceniHrany = nastaveni.OhodnoceniHrany;
    }

    /// <summary>
    /// Načte graf ze souboru a vykreslí jej.
    /// </summary>
    /// <param name="cestaNazev">Cesta k souboru a jeho název.</param>
    public void NactiGrafBinarne(string cestaNazev)
    {
      GGraf novyGraf = BinarniDatovyModul.GetInstance.NactiGraf(cestaNazev);
      if (novyGraf != null)
        Graf = novyGraf;
      else
        throw new ArgumentException("Soubor " + cestaNazev + " není platným souborem grafu.");
    }

    /// <summary>
    /// Uloží graf do souboru.
    /// <exception cref="System.ArgumentException">Metoda vyvolá výjimku, pokud soubor nelze vytvořit,
    /// nebo pokud graf nelze serializovat.</exception>
    /// </summary>
    /// <param name="cestaNazev">Cesta k souboru a jeho název.</param>
    public void UlozGrafBinarne(string cestaNazev)
    {
      BinarniDatovyModul.GetInstance.UlozGraf(cestaNazev, Graf);
    }

    /// <summary>
    /// Uloží graf požadovaným způsobem.
    /// </summary>
    /// <param name="modul">Způsob uložení grafu ve formě callback metody.</param>
    /// <param name="cestaNazev">Cesta k souboru a jeho název.</param>
    public void UlozGraf(IDatovyModul modul, string cestaNazev)
    {
      modul.UlozGraf(cestaNazev, Graf);
    }

    /// <summary>
    /// Načte graf požadovaným způsobem.
    /// </summary>
    /// <param name="modul">Způsob načtení grafu ve formě callback metody.</param>
    /// <param name="cestaNazev">Cesta k souboru a jeho název.</param>
    public void NactiGraf(IDatovyModul modul, string cestaNazev)
    {
      GGraf g = modul.NactiGraf(cestaNazev);
      if (g != null )
        Graf = g;
      else
        throw new ArgumentException("Soubor " + cestaNazev + " není platným souborem grafu.");
      
    }
    #endregion


    #region Sledované události
    /// <summary>
    /// Provede celkové překreslení kreslící plochy.
    /// </summary>
    /// <param name="e">Doplňující informace</param>
    protected override void OnPaint(PaintEventArgs e)
    {
      kpPlatno.Invalidate();
      // kpPlatno_Paint(this, e);
    }

    private void kpPlatno_Paint(object sender, PaintEventArgs e)
    {
      base.OnPaint(e);

      using (Graphics g = kpPlatno.CreateGraphics())
      {
        g.Clear(this.BackColor);
        VykresliOsu(g);
        VykresliGraf(g);
      }
    }
    // Myš
    private void kpPlatno_MouseMove(object sender, MouseEventArgs e)
    {
      OnMouseMove(e);

      // Aktualizace souřadnic kurzoru.
      Bod<int> souradniceKurzoruPixelove = new Bod<int>(e.X, e.Y);
      Bod<double> souradniceKurzoruRealne = ss.P2R(souradniceKurzoruPixelove);
      souradniceKurzoru.X = (int)Math.Round(souradniceKurzoruRealne.X);
      souradniceKurzoru.Y = (int)Math.Round(souradniceKurzoruRealne.Y);

      // Najdu aktuální vrchol/hranu.
      NajdiAktualniPrvek(new Bod<double>(Math.Round(souradniceKurzoruRealne.X, 0),
        Math.Round(souradniceKurzoruRealne.Y, 0)));

      // Vykreslím barvy.
      if (!rezRoll && VykreslovaniAktivnihoPrvku && aktualniPrvek != null)
      {
        if (preAktualniPrvek == null || preAktualniPrvek != aktualniPrvek)
        {
          using (Graphics g = kpPlatno.CreateGraphics())
          {
            if (rezimPrace == TypRezmimuPrace.PraceSVrcholy)
            {
              (aktualniPrvek as GVrchol).JeAktivni = true;
              VykresliVrchol(aktualniPrvek as GVrchol, g, true);
            }
            else
            {
              (aktualniPrvek as GHrana).JeAktivni = true;
              VykresliHranu(aktualniPrvek as GHrana, g, true);
            }

            if (preAktualniPrvek is GVrchol)
            {
              (preAktualniPrvek as GVrchol).JeAktivni = false;
              VykresliVrchol(preAktualniPrvek as GVrchol, g, true);
            }
            else if (preAktualniPrvek is GHrana)
            {
              (preAktualniPrvek as GHrana).JeAktivni = false;
              VykresliHranu(preAktualniPrvek as GHrana, g, true);
            }
          }
          preAktualniPrvek = aktualniPrvek;
        }
      }

      // Pokud se roluje:
      if (rezRoll)
      {
        ss.NavigacePosunPX(rollovaciSouradnice.X - e.X, -(rollovaciSouradnice.Y - e.Y));
        rollovaciSouradnice.X = e.X;
        rollovaciSouradnice.Y = e.Y;
        kpPlatno.Invalidate();
      }
    }

    private void kpPlatno_MouseClick(object sender, MouseEventArgs e)
    {
      // přidání vrcholu
      if (e.Button == System.Windows.Forms.MouseButtons.Left && Control.ModifierKeys == controlPridavaciRezim && 
        Control.ModifierKeys != controlOdstranovaciRezim && rezimPrace == TypRezmimuPrace.PraceSVrcholy &&
        Control.ModifierKeys != controlEditacniRezim && !rezPridHra && GrafickaEditaceStruktury)
      {
        Bod<int> bodP = new Bod<int>(e.X, e.Y);
        GVrchol vrchol = new GVrchol(new SouradniceVrcholu(Math.Round(ss.P2R(bodP).X), Math.Round(ss.P2R(bodP).Y)));
        vrchol.BarvaOhraniceni = VnejsiBarvaVrcholu;
        vrchol.BarvaTextu = BarvaTextuVrcholu;
        vrchol.BarvaVyplne = VnitrniBarvaVrcholu;
        vrchol.Velikost = VelikostVrcholu;
        vrchol.ZobrazitText = ZobrazitTextVrcholu;
        vrchol.FontTextu = FontVrcholu;
        try
        {
          graf.VlozVrchol(vrchol);
          // odpálím informační event vnějšku
          ZmenaStrukturyGrafuEventArgs ev = new ZmenaStrukturyGrafuEventArgs(vrchol, TypZmenyStrukturyGrafu.VlozeniVrcholu);
          OnZmenaStrukturyGrafu(this, ev);
        }
        catch (Exception ex)
        {
          MessageBox.Show(ex.Message, "Chyba aplikace", MessageBoxButtons.OK, MessageBoxIcon.Error);
        }
      }

      // zkratka pro přidání hrany, pokud druhý klik mine, akce se ruší.
      if (e.Button == System.Windows.Forms.MouseButtons.Right && Control.ModifierKeys == controlPridavaciRezim 
        && Control.ModifierKeys != controlEditacniRezim && rezimPrace == TypRezmimuPrace.PraceSVrcholy
        && Control.ModifierKeys != controlOdstranovaciRezim && GrafickaEditaceStruktury)
      {
        if (!rezPridHra)
        {
          iv1 = aktualniPrvek as GVrchol;
          rezPridHra = true;
          kpPlatno.Cursor = Cursors.Cross;
          return;
        }
        else
        {
          iv2 = aktualniPrvek as GVrchol;
        }

        if ((iv1 != null && iv2 != null) && !ReferenceEquals(iv1, iv2))
        {
          bool nasobna = false;
          foreach (GHrana gh in iv1.SeznamSousedu)
          {
            if (gh.IncidencniVrchol1 == iv2 || gh.IncidencniVrchol2 == iv2)
            {
              nasobna = true;
            }
          }
          if (!nasobna)
          {
            GHrana hrana = new GHrana(iv1, iv2);
            hrana.Tloustka = TloustkaHrany;
            hrana.Ohodnoceni = OhodnoceniHrany;
            hrana.ZobrazitText = ZobrazitTextHrany;
            hrana.Barva = BarvaHrany;
            hrana.BarvaTextu = BarvaTextuHrany;
            hrana.FontTextu = FontHrany;
            try
            {
              graf.VlozHranu(hrana, hrana.Klic);
              rezPridHra = false;
              iv1 = null;
              iv2 = null;
              // odpálím informační event vnějšku
              ZmenaStrukturyGrafuEventArgs ev = new ZmenaStrukturyGrafuEventArgs(
                graf.NajdiHranu(hrana.Klic), TypZmenyStrukturyGrafu.VlozeniHrany
              );
              OnZmenaStrukturyGrafu(this, ev);
            }
            catch (Exception ex)
            {
              MessageBox.Show(ex.Message, "Chyba aplikace", MessageBoxButtons.OK, MessageBoxIcon.Error);
              rezPridHra = false;
              iv1 = null;
              iv2 = null;
            }
          }
          else
          {
            MessageBox.Show("Nelze přidávat násobné hrany", "Chyba aplikace", MessageBoxButtons.OK, MessageBoxIcon.Error);
            rezPridHra = false;
            iv1 = null;
            iv2 = null;
          }
          kpPlatno.Cursor = Cursors.Default;
        }
        else
        {
          iv1 = null;
          iv2 = null;
          rezPridHra = false;
          kpPlatno.Cursor = Cursors.Default;
        }
      }

      // zkratka pro editační režim na aktualním vrcholu/hraně 
      if (e.Button == System.Windows.Forms.MouseButtons.Left && Control.ModifierKeys == controlEditacniRezim && Control.ModifierKeys 
        != controlPridavaciRezim && aktualniPrvek != null &&
        Control.ModifierKeys != controlOdstranovaciRezim && !rezPridHra && GrafickaEditaceStruktury)
      {
        if (aktualniPrvek is GVrchol)
        {
          // Změním
          UpravVrchol(aktualniPrvek as GVrchol);
          // odpálím informační event vnějšku
          ZmenaStrukturyGrafuEventArgs ev = 
              new ZmenaStrukturyGrafuEventArgs(aktualniPrvek, TypZmenyStrukturyGrafu.EditaceVrcholu);
          OnZmenaStrukturyGrafu(this, ev);
        }
        else
        {
          // Změním
          UpravHranu(aktualniPrvek as GHrana);
          // odpálím informační event vnějšku
          ZmenaStrukturyGrafuEventArgs ev = 
              new ZmenaStrukturyGrafuEventArgs(aktualniPrvek, TypZmenyStrukturyGrafu.EditaceHrany);
          OnZmenaStrukturyGrafu(this, ev);
        }
      }

      // zkratka pro odstranění aktivního vrcholu/hrany stm 
      if ((e.Button == System.Windows.Forms.MouseButtons.Left && Control.ModifierKeys == controlOdstranovaciRezim && 
        Control.ModifierKeys != controlEditacniRezim && aktualniPrvek != null &&
        Control.ModifierKeys != controlPridavaciRezim && !rezPridHra && GrafickaEditaceStruktury))
      {
        if (aktualniPrvek is GVrchol)
        {
          // pokud vrchol neinciduje
          if ((aktualniPrvek as GVrchol).Stupen == 0)
          {
            // odpálím cancel event
            ZmenaStrukturyGrafuCancelEventArgs ec = 
              new ZmenaStrukturyGrafuCancelEventArgs(aktualniPrvek);
            OnOdstraneniPrvku(this, ec);
            if (!ec.Cancel)
            {
              GVrchol vrchol = 
                graf.OdstranVrchol((aktualniPrvek as GVrchol).Klic);
              // odpálím informační event vnějšku
              ZmenaStrukturyGrafuEventArgs ev = 
                new ZmenaStrukturyGrafuEventArgs(vrchol, TypZmenyStrukturyGrafu.OdstraneniVrcholu);

              OnZmenaStrukturyGrafu(this, ev);
            }
          }
        }
        else
        {
          // odpálím cancel event
          ZmenaStrukturyGrafuCancelEventArgs ec = 
            new ZmenaStrukturyGrafuCancelEventArgs(aktualniPrvek);
          OnOdstraneniPrvku(this, ec);
          if (!ec.Cancel)
          {
            GHrana hrana = graf.OdstranHranu((aktualniPrvek as GHrana).Klic);
            // odpálím informační event vnějšku
            ZmenaStrukturyGrafuEventArgs ev = 
              new ZmenaStrukturyGrafuEventArgs(hrana, TypZmenyStrukturyGrafu.OdstraneniHrany);
            OnZmenaStrukturyGrafu(this, ev);
          }
        }
      }
    }

    private void kpPlatno_MouseDown(object sender, MouseEventArgs e)
    {
      if (e.Button == System.Windows.Forms.MouseButtons.Left)
      {
        rezRoll = true;
        rollovaciSouradnice.X = e.X;
        rollovaciSouradnice.Y = e.Y;
        kpPlatno.Cursor = Cursors.SizeAll;
      }
    }

    private void kpPlatno_MouseUp(object sender, MouseEventArgs e)
    {
      if (e.Button == System.Windows.Forms.MouseButtons.Left && rezRoll)
      {
        rezRoll = false;
        if (rezPridHra)
          kpPlatno.Cursor = Cursors.Cross;
        else
          kpPlatno.Cursor = Cursors.Default;
      }
    }

    /// <summary>
    /// Nastane při otočení kolečka myši.
    /// </summary>
    /// <param name="e">Doplňující informace</param>
    /// <param name="sender">Odesílatel eventu.</param>
    private void kpPlatno_MouseWheel(object sender, MouseEventArgs e)
    {
      base.OnMouseWheel(e);
      ss.NavigaceZoom(-(e.Delta / 10));
      kpPlatno.Invalidate();
    }

    // tlačítka
    /// <summary>
    /// Nastane při stlačení tlačítka na klávesnici.
    /// </summary>
    /// <param name="e">Doplňující informace.</param>
    /// <param name="sender">Odesílatel eventu.</param>
    private void kpPlatno_KeyDown(object sender, KeyEventArgs e)
    {
      base.OnKeyDown(e);
      bool zmena = false;

      if (e.KeyCode == Keys.PageDown)
      {
        ss.NavigaceZoom(4);
        e.Handled = true;
        zmena = true;
      }

      if (e.KeyCode == Keys.PageUp)
      {
        ss.NavigaceZoom(-4);
        e.Handled = true;
        zmena = true;
      }

      if (e.KeyCode == Keys.Up)
      {
        ss.NavigacePosun(0, -15);
        e.Handled = true;
        zmena = true;
      }

      if (e.KeyCode == Keys.Left)
      {
        ss.NavigacePosun(15, 0);
        e.Handled = true;
        zmena = true;
      }

      if (e.KeyCode == Keys.Down)
      {
        ss.NavigacePosun(0, 15);
        e.Handled = true;
        zmena = true;
      }

      if (e.KeyCode == Keys.Right)
      {
        ss.NavigacePosun(-15, 0);
        e.Handled = true;
        zmena = true;
      }

      if (zmena)
        kpPlatno.Invalidate();
    }

    // ostatní
    private void kpPlatno_Resize(object sender, EventArgs e)
    {
      ss.RozpetiPX = kpPlatno.Width;
      ss.RozpetiPY = kpPlatno.Height;
      kpPlatno.Invalidate();
    }

    private void pbNastaveni_Click(object sender, EventArgs e)
    {
      using (FrmNastaveni frmNastaveni = new FrmNastaveni(NaplnPrepravku()))
      {
        if (frmNastaveni.ShowDialog() == System.Windows.Forms.DialogResult.OK)
        {
          ZmenNastaveni(frmNastaveni.Nastaveni);
          if (frmNastaveni.Reset)
          {
            foreach (GVrchol item in Graf)
            {
              item.BarvaOhraniceni = VnejsiBarvaVrcholu;
              item.BarvaVyplne = VnitrniBarvaVrcholu;
              item.BarvaTextu = BarvaTextuVrcholu;
              item.FontTextu = FontVrcholu;
              item.Velikost = VelikostVrcholu;
              item.ZobrazitText = ZobrazitTextVrcholu;
            }
            foreach (IPrvekTabulky<GHrana, double> item in Graf.SeznamHran)
            {
              GHrana hrana = item.Data;
              hrana.Barva = BarvaHrany;
              hrana.BarvaTextu = BarvaTextuHrany;
              hrana.FontTextu = FontHrany;
              hrana.Tloustka = TloustkaHrany;
              hrana.ZobrazitText = ZobrazitTextHrany;
            }
          }
        }
      }
    }

    private void pbInformace_Click(object sender, EventArgs e)
    {
      Process p = new Process();
      p.StartInfo.FileName ="iexplore.exe";
      p.StartInfo.Arguments = Directory.GetCurrentDirectory() + "/informace/informace.html";
      p.Start();
    }

    private void pbTypEditace_Click(object sender, EventArgs e)
    {
      if (rezimPrace == TypRezmimuPrace.PraceSVrcholy)
      {
        pbTypEditace.BackgroundImage = Properties.Resources.editace_hran;
        rezimPrace = TypRezmimuPrace.PraceSHranami;
      }
      else
      {
        pbTypEditace.BackgroundImage = Properties.Resources.editace_vrcholu;
        rezimPrace = TypRezmimuPrace.PraceSVrcholy;
      }
    }
    #endregion


    #region Konstruktor
    /// <summary>
    /// Bezparametrický konsturktor, inicializuje potřebné datové složky a alokuje složené objekty.
    /// </summary>
    public ObycejnyGraf()
    {
      InitializeComponent();

      graf = new GGraf(this);
      ss = new SouradnySystem2D();
      souradniceKurzoru = new Bod<int>();

      ss.RozpetiPX = kpPlatno.Width;
      ss.RozpetiPY = kpPlatno.Height;
      ss.RealMinX = -40;
      ss.RealMaxX = 40;
      ss.RealMinY = -20;
      ss.RealMaxY = 20;

      AutomatickeUkladaniNacitaniKonfigurace = true;
      OhodnoceniHrany = 1;

      kpPlatno.BackColor = Color.White;
      BarvaOsy = Color.Gray;
      VnejsiBarvaVrcholu = Color.Blue;
      VnitrniBarvaVrcholu = Color.White;
      BarvaHrany = Color.Black;
      BarvaTextuVrcholu = Color.Black;
      BarvaTextuHrany = Color.Black;
      BarvaAktivnihoPrvku = Color.Red;
      VelikostVrcholu = 10;
      FontVrcholu = new Font("arial", 9);
      ZobrazitTextVrcholu = true;
      TloustkaHrany = 3;
      FontHrany = new Font("arial", 9);
      ZobrazitTextHrany = true;
    }
    #endregion
  }
}