﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using MG.DatoveStruktury.Tabulky;

namespace MG.DatoveStruktury.Grafy.SmisenaReprezentace
{
  /// <summary>
  /// Smíšená reprezentace neorientovaného grafu.
  /// </summary>
  /// <typeparam name="TVrchol">Datový typ reprezentující vrchol.</typeparam>
  /// <typeparam name="THrana">Datový typ reprezentující hranu.</typeparam>
  /// <typeparam name="TKlicVrcholu">Datový typ, který je klíčem vrcholů.</typeparam>
  /// <typeparam name="TKlicHrany">Datový typ, který je klíčem hran.</typeparam>
  /// <typeparam name="TOhodnoceniHrany">Datový typ, který je ohodnocením hran.</typeparam>
  [Serializable]
  public class AdsGrafSmisenaReprezentace<TVrchol, THrana, TKlicVrcholu, TKlicHrany, TOhodnoceniHrany> :
    IEnumerable<TVrchol>
    where TVrchol : Vrchol<TKlicVrcholu, TKlicHrany, TOhodnoceniHrany>
    where THrana : Hrana<TKlicHrany, TKlicVrcholu, TOhodnoceniHrany>
    where TKlicVrcholu : IComparable<TKlicVrcholu>
    where TKlicHrany : IComparable<TKlicHrany>
  {
    #region Datové složky
    /// <summary>
    /// Primární datová struktura. Využita struktura binární vyhledávací strom (tabulka).
    /// </summary>
    protected AdsBinarniVyhledavaciStrom<TVrchol, TKlicVrcholu> primarniDatovaStruktura = 
      new AdsBinarniVyhledavaciStrom<TVrchol, TKlicVrcholu>();

    /// <summary>
    /// Sekundární datová sturktura. Seznam hran grafu implementovaný utříděnou tabulkou na seznamu.
    /// </summary>
    protected AdsTabulkaSeznam<THrana, TKlicHrany> sekundarniDatovaStruktura = 
      new AdsTabulkaSeznam<THrana, TKlicHrany>();
    #endregion

    #region Vlastnosti
    /// <summary>
    /// Počet vrcholů grafu.
    /// </summary> 
    public int PocetVrcholu
    {
      get { return primarniDatovaStruktura.Mohutnost; }
    }

    /// <summary>
    /// Počet hran grafu.
    /// </summary>
    public int PocetHran
    {
      get { return sekundarniDatovaStruktura.Mohutnost; }
    }

    /// <summary>
    /// Seznam hran grafu. Seznam je realizovaný utříděnou tabulkou na seznamu.
    /// </summary>
    public virtual AdsTabulkaSeznam<THrana, TKlicHrany> SeznamHran
    {
      get { return sekundarniDatovaStruktura; }
    }
    #endregion

    #region Metody
    /// <summary>
    /// Vloží vrchol do struktury grafu.
    /// </summary>
    /// <param name="vrchol">Vrchol , který se vloží do grafu.</param>
    /// <param name="klic">Klíč vrcholu.</param>
    /// <exception cref="System.ArgumentException">Metoda je chráněna výjimkou pro případ, že by se přidával duplicitní vrchol.</exception>
    public virtual void VlozVrchol(TVrchol vrchol, TKlicVrcholu klic)
    {
      try { primarniDatovaStruktura.VlozPrvek(vrchol, klic); }
      catch (Exception)
      {
        throw new ArgumentException("Vrchol na daných souřadnicích již existuje", "klic");
      }
    }

    /// <summary>
    /// Vyhledá vrchol ve struktuře grafu podle klíče.
    /// </summary>
    /// <param name="klic">Klíč vrcholu.</param>
    /// <returns>Vrchol dle klíče nebo hodnota <c>null</c>.</returns>
    public virtual TVrchol NajdiVrchol(TKlicVrcholu klic)
    {
      return primarniDatovaStruktura.NajdiData(klic);
    }

    /// <summary>
    /// Vyhledá vrchol dle klíče, následně se ho pokusí odstranit.
    /// <remarks>Metoda nekontroluje, zda stupeň vrcholu je 1.</remarks>
    /// </summary>
    /// <param name="klic">Klíč odstraňovaného vrcholu.</param>
    /// <returns>Odstraněný vrchol nebo hodnota <c>null</c> v případě, 
    /// že daný vrchol neexistuje.</returns>
    public virtual TVrchol OdstranVrchol(TKlicVrcholu klic)
    {
      return primarniDatovaStruktura.OdstranPrvek(klic);
    }

    /// <summary>
    /// Vloží hranu do grafu.
    /// </summary>
    /// <param name="iv1">Klíč prvního incidenčního vrcholu.</param>
    /// <param name="iv2">Klíč druhého incidenčního vrcholu.</param>
    /// <param name="hrana">Vkáladná hrana.</param>
    /// <param name="klic">Klíč hrany.</param>
    /// <exception cref="System.ArgumentException">Metoda vyvolá výjimku v případě, že 
    /// některý z incidenčních vrcholů není součástí množiny vrcholů grafu.</exception>
    public virtual void VlozHranu(TKlicVrcholu iv1, TKlicVrcholu iv2,
      THrana hrana, TKlicHrany klic)
    {
      Vrchol<TKlicVrcholu, TKlicHrany, TOhodnoceniHrany> v1 = primarniDatovaStruktura.NajdiData(iv1);
      Vrchol<TKlicVrcholu, TKlicHrany, TOhodnoceniHrany> v2 = primarniDatovaStruktura.NajdiData(iv2);
      if (v1 == null || v2 == null)
      {
        throw new ArgumentException("Některý z incidenčních vrcholů hrany není součástí množiny vrcholů  grafu.");
      }

      sekundarniDatovaStruktura.VlozPrvek(hrana, klic);
      hrana.IncidencniVrchol1.VlozSouseda(hrana);
      hrana.IncidencniVrchol2.VlozSouseda(hrana);
    }

    /// <summary>
    /// Nalezne hranu dle zadaného klíče.
    /// </summary>
    /// <param name="klic">Klíč hledané hrany.</param>
    /// <returns>Hledanou hranu nebo <c>null</c>.</returns>
    public virtual THrana NajdiHranu(TKlicHrany klic)
    {
      return sekundarniDatovaStruktura.NajdiData(klic);
    }

    /// <summary>
    /// Odstraní hranu z grafu.
    /// </summary>
    /// <param name="klic">Klíč odstraňované hrany.</param>
    /// <returns>Odstraňenou hranu, nebo <c>null</c>.</returns>
    public virtual THrana OdstranHranu(TKlicHrany klic)
    {
      THrana hrana = sekundarniDatovaStruktura.OdeberPrvek(klic);

      if (hrana != null)
      {
        hrana.IncidencniVrchol1.OdeberSouseda(hrana);
        hrana.IncidencniVrchol2.OdeberSouseda(hrana);
      }

      return hrana;
    }

    /// <summary>
    /// Prohlídka vrcholů do hloubky. Vrcholy jsou zpřístupňovány vzestupně dle hodnoty klíče.
    /// </summary>
    /// <returns>IEnumerator pro cyklus foreach.</returns>
    public virtual IEnumerator<TVrchol> GetEnumerator()
    {
      foreach (IPrvekTabulky<TVrchol, TKlicVrcholu> vrchol in primarniDatovaStruktura)
      {
        yield return vrchol.Data;
      }
    }

    System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
    {
      return GetEnumerator();
    }
    #endregion

    #region Konstruktor
    #endregion
  }
}
