﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace MG.DatoveStruktury.Grafy.SmisenaReprezentace{
  /// <summary>
  /// Hrana grafu.
  /// </summary>
  /// <typeparam name="TKlicHrany">Referenční datový typ, který je klíčem hrany. 
  /// Typ musí implementovat rozhraní <see cref="IComparable"/>.</typeparam>
  /// <typeparam name="TKlicVrcholu">Referenční datový typ, který je klíčem incidenčních vrcholů. 
  /// Typ musí implementovat rozhraní <see cref="IComparable"/>.</typeparam>
  /// <typeparam name="TOhodnoceniHrany">Referenční datový typ, který je ohodnocením hrany.</typeparam>
  [Serializable]
  public class Hrana<TKlicHrany, TKlicVrcholu, TOhodnoceniHrany> : IPrvekGrafu
    where TKlicHrany : IComparable<TKlicHrany>
    where TKlicVrcholu : IComparable<TKlicVrcholu>
  {
    #region Vlastnosti
    /// <summary>
    /// První incidenční vrchol hrany.
    /// </summary>
    public virtual Vrchol<TKlicVrcholu, TKlicHrany, TOhodnoceniHrany> IncidencniVrchol1 { get; private set; }

    /// <summary>
    /// Druhý incidenční vrchol hrany.
    /// </summary>
    public virtual Vrchol<TKlicVrcholu, TKlicHrany, TOhodnoceniHrany> IncidencniVrchol2 { get; private set; }

    /// <summary>
    /// Reference na dalšího souseda prvního incidenčního vrcholu.
    /// </summary>
    internal Hrana<TKlicHrany, TKlicVrcholu, TOhodnoceniHrany> SousedIV1 { get; private set; }

    /// <summary>
    /// Reference na dalšího souseda druhého incidenčního vrcholu.
    /// </summary>
    internal Hrana<TKlicHrany, TKlicVrcholu, TOhodnoceniHrany> SousedIV2 { get; private set; }

    /// <summary>
    /// Ohodnocení hrany.
    /// </summary>
    public virtual TOhodnoceniHrany Ohodnoceni { get; set; }

    /// <summary>
    /// Doplňková informace.
    /// </summary>
    public virtual object Tag { get; set; }
    #endregion

    #region Metody
    /// <summary>
    /// Vloží souseda do možiny sousedů daného vrcholu.
    /// </summary>
    /// <param name="sender">Vrchol, který žádá o zavedení souseda.</param>
    /// <param name="hrana">Hrana, která se má evidovat jako soused.</param>
    internal void VlozSouseda(Vrchol<TKlicVrcholu, TKlicHrany, TOhodnoceniHrany> sender,
      Hrana<TKlicHrany, TKlicVrcholu, TOhodnoceniHrany> hrana)
    {
      if (ReferenceEquals(sender, IncidencniVrchol1))
      {
        SousedIV1 = hrana;
      }
      else
      {
        SousedIV2 = hrana;
      }
    }

    /// <summary>
    /// Zpřístupní následujícího souseda daného vrcholu.
    /// </summary>
    /// <param name="sender">Vrchol, který žádá  o zpřístupnění</param>
    /// <returns>Následujícího souseda nebo <c>null</c>.</returns>
    internal Hrana<TKlicHrany, TKlicVrcholu, TOhodnoceniHrany> 
      DejSouseda(Vrchol<TKlicVrcholu, TKlicHrany, TOhodnoceniHrany> sender)
    {
      if (ReferenceEquals(sender, IncidencniVrchol1))
        return SousedIV1;
      else
        return SousedIV2;
    }
    #endregion

    /// <summary>
    /// Parametrický konstruktor.
    /// </summary>
    /// <param name="iv1">První incidenční vrchol hrany.</param>
    /// <param name="iv2">Druhý incidenční vrchol hrany.</param>
    public Hrana(Vrchol<TKlicVrcholu, TKlicHrany, TOhodnoceniHrany> iv1,
      Vrchol<TKlicVrcholu, TKlicHrany, TOhodnoceniHrany> iv2)
    {
      IncidencniVrchol1 = iv1;
      IncidencniVrchol2 = iv2;
    }
  }
}
