﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using MG.DatoveStruktury.LinearniMnoziny;

namespace MG.DatoveStruktury.Grafy.SmisenaReprezentace
{
  /// <summary>
  /// Vrchol grafu.
  /// </summary>
  /// <typeparam name="TKlicVrcholu">Referenční datový typ, který je klíčem vrcholu. 
  /// Typ musí implementovat rozhraní <see cref="IComparable"/>.</typeparam>
  /// <typeparam name="TKlicHrany">Referenční datový typ, který je klíčem incidenčních hran. 
  /// Typ musí implementovat rozhraní <see cref="IComparable"/>.</typeparam>
  /// <typeparam name="TOhodnoceniHrany">Referenční datový typ, který je ohodnocením hrany.</typeparam>
  [Serializable]
  public class Vrchol<TKlicVrcholu, TKlicHrany, TOhodnoceniHrany> : IPrvekGrafu
    where TKlicVrcholu : IComparable<TKlicVrcholu>
    where TKlicHrany : IComparable<TKlicHrany>
  {
    #region Datové složky
    private Hrana<TKlicHrany, TKlicVrcholu, TOhodnoceniHrany> pocatecniSoused = null;
    private Hrana<TKlicHrany, TKlicVrcholu, TOhodnoceniHrany> posledniSoused = null;
    int stupen = 0;
    bool synchronizovany = true;
    private AdsObousmernySeznam<Hrana<TKlicHrany, TKlicVrcholu, TOhodnoceniHrany>> seznam =
      new AdsObousmernySeznam<Hrana<TKlicHrany,TKlicVrcholu,TOhodnoceniHrany>>();
    #endregion

    #region Vlastnosti
    /// <summary>
    /// Doplňková informace.
    /// </summary>
    public virtual object Tag { get; set; }

    /// <summary>
    /// Stupeň vrcholu.
    /// </summary>
    public int Stupen { get { return stupen; } }

    /// <summary>
    /// Seznam sousedních hran vrcholu. Při prvním volání po změně okolí vrcholu se generuje, pro další
    /// použití je uložen až do další editace okolí vrcholu.
    /// </summary>
    public AdsObousmernySeznam<Hrana<TKlicHrany, TKlicVrcholu, TOhodnoceniHrany>>
      SeznamSousedu { get { return GetSeznamSousedu(); } }
    #endregion

    #region Metody
    private AdsObousmernySeznam<Hrana<TKlicHrany, TKlicVrcholu, TOhodnoceniHrany>> GetSeznamSousedu()
    {
      if (!synchronizovany)
      {
        seznam = new AdsObousmernySeznam<Hrana<TKlicHrany, TKlicVrcholu, TOhodnoceniHrany>>();
        Hrana<TKlicHrany, TKlicVrcholu, TOhodnoceniHrany> h = pocatecniSoused;
        while (h != null)
        {
          seznam.VlozPosledni(h);
          h = h.DejSouseda(this);
        }

        synchronizovany = true;
        return seznam;
      }
      else
      {
        return seznam;
      }
    }

    /// <summary>
    /// Zaeviduje hranu jako svého souseda.
    /// </summary>
    /// <param name="hrana">Hrana, která se má zaevidovat.</param>
    internal void VlozSouseda(Hrana<TKlicHrany, TKlicVrcholu, TOhodnoceniHrany> hrana)
    {
      stupen++;
      synchronizovany = false;
      if (pocatecniSoused == null)
      {
        pocatecniSoused = hrana;
        posledniSoused = pocatecniSoused;
      }
      else
      {
        posledniSoused.VlozSouseda(this, hrana);
        posledniSoused = hrana;
      }
    }

    /// <summary>
    /// Odebere hranu ze svého seznamu sousedů.
    /// </summary>
    /// <param name="hrana">Odebíraná hrana.</param>
    internal void OdeberSouseda(Hrana<TKlicHrany, TKlicVrcholu, TOhodnoceniHrany> hrana)
    {
      stupen--;
      synchronizovany = false;
      Hrana<TKlicHrany, TKlicVrcholu, TOhodnoceniHrany> aktualni = null;
      // Odebírá se první soused
      if (ReferenceEquals(pocatecniSoused, hrana))
      {
        aktualni = pocatecniSoused.DejSouseda(this);
        // Je jich víc než jen ten první
        if (aktualni != null)
        {
          pocatecniSoused = aktualni;
          return;
        } // Je jen jeden soused
        else
        {
          pocatecniSoused = null;
          posledniSoused = null;
          return;
        }
      }

      Hrana<TKlicHrany, TKlicVrcholu, TOhodnoceniHrany> preAktualni = pocatecniSoused;
      aktualni = preAktualni.DejSouseda(this);
      do
      {
        // Pokud je toto odebíráný soused
        if (ReferenceEquals(hrana, aktualni))
        {// Pokud je odebíraný soused poslední soused
          if (ReferenceEquals(aktualni, posledniSoused))
          {
            posledniSoused = preAktualni;
            preAktualni.VlozSouseda(this, null);
          }
          else
          {
            aktualni = aktualni.DejSouseda(this);
            preAktualni.VlozSouseda(this, aktualni);
            break;
          }
        }

        preAktualni = aktualni;
        aktualni = aktualni.DejSouseda(this);
      } while (aktualni!= null);
    }
    #endregion
  }
}