﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace MG.DatoveStruktury.LinearniMnoziny
{
  /// <summary>
  /// Abstraktní datová struktura Obousměrný necyklický lineární seznam.
  /// <remarks>Složitost všech operací je O(1).</remarks>
  /// </summary>
  /// <typeparam name="T">Referenční datový typ, který obsahují prvky seznamu.</typeparam>
  [Serializable]
  public class AdsObousmernySeznam<T> : IEnumerable<T>
  {
    #region Interní datové typy
    [Serializable]
    class PrvekObousmerehoSeznamu
    {
      #region Vlastnosti
      public PrvekObousmerehoSeznamu Predchudce {get; set;}

      public PrvekObousmerehoSeznamu Naslednik {get; set;}

      public T data { get; set; }
      #endregion


      #region Konstruktor
      internal PrvekObousmerehoSeznamu(T data)
      {
        this.data = data;
        this.Predchudce = null;
        this.Naslednik = null;
      }
      #endregion
    }
    #endregion


    #region Datové složky
    private PrvekObousmerehoSeznamu prvni;

    private PrvekObousmerehoSeznamu posledni;

    private PrvekObousmerehoSeznamu aktualni;
    #endregion


    #region Metody
    /// <summary>
    /// <c>True</c> je-li seznam prázdný, <c>False</c> jinak.
    /// </summary>
    public bool JePrazdny
    {
      get
      {
        return prvni == null;
      }
    }

    /// <summary>
    /// Vloží prvek s danými daty na začátek seznamu.
    /// </summary>
    /// <param name="data">Data, která má prvek obsahovat.</param>
    public void VlozPrvni(T data)
    {
      if (JePrazdny)
      {
        prvni = new PrvekObousmerehoSeznamu(data);
        aktualni = prvni;
        posledni = aktualni;
      }
      else
      {
        PrvekObousmerehoSeznamu pom = new PrvekObousmerehoSeznamu(data);
        prvni.Predchudce = pom;
        pom.Naslednik = prvni;
        prvni = pom;
      }
    }

    /// <summary>
    /// Vloží prvek s danými daty na konec seznamu.
    /// </summary>
    /// <param name="data">Data, která má prvek obsahovat.</param>
    public void VlozPosledni(T data)
    {
      if (JePrazdny)
        VlozPrvni(data);
      else
      {
        PrvekObousmerehoSeznamu pom = new PrvekObousmerehoSeznamu(data);
        posledni.Naslednik = pom;
        pom.Predchudce = posledni;
        posledni = pom;
      }
    }

    /// <summary>
    /// Vloží prvek s danými daty za aktuální prvek seznamu.
    /// </summary>
    /// <param name="data">Data, která má prvek obsahovat.</param>
    public void VlozNaslednika(T data)
    {
      if (JePrazdny)
        VlozPrvni(data);
      else
        if (aktualni == posledni)
          VlozPosledni(data);
        else
        {
          PrvekObousmerehoSeznamu pom2 = aktualni.Naslednik;
          PrvekObousmerehoSeznamu pom1 = new PrvekObousmerehoSeznamu(data);
          pom1.Naslednik = pom2;
          pom2.Predchudce = pom1;
          pom1.Predchudce = aktualni;
          aktualni.Naslednik = pom1;
        }
    }

    /// <summary>
    /// Vloží prvek s danými daty před aktuální prvek seznamu.
    /// </summary>
    /// <param name="data">Data, která má prvek obsahovat.</param>
    public void VlozPredchudce(T data)
    {
      if (JePrazdny || (aktualni == prvni))
        VlozPrvni(data);
      else
      {
        PrvekObousmerehoSeznamu pom2 = aktualni.Predchudce;
        PrvekObousmerehoSeznamu pom1 = new PrvekObousmerehoSeznamu(data);
        pom1.Predchudce = pom2;
        pom2.Naslednik = pom1;
        pom1.Naslednik = aktualni;
        aktualni.Predchudce = pom1;
      }
    }

    /// <summary>
    /// Odebere první prvek seznamu.
    /// <remarks>Pokud je aktuální prvek první prvek, reference aktuálního prvku
    /// se přesune na následníka.</remarks>
    /// </summary>
    /// <returns>Default(<typeparamref name="T"/>) pokud je seznam prázdný, 
    /// jinak data typu <typeparamref name="T"/>.</returns>
    public T OdeberPrvni()
    {
      if (JePrazdny)
        return default(T);

      PrvekObousmerehoSeznamu pom1 = prvni;
      if (aktualni == prvni)
      {
        if (aktualni == posledni)
          posledni = posledni.Naslednik;

        aktualni = prvni.Naslednik;
        prvni = aktualni;
      }
      else
        prvni = prvni.Naslednik;

      if (prvni != null)
        prvni.Predchudce = null;
      return pom1.data;
    }

    /// <summary>
    /// Odebere poslední prvek seznamu.
    /// <remarks>Pokud je aktuální prvek poslední prvek, reference aktuálního prvku 
    /// se přenese na předchůdce.</remarks>
    /// </summary>
    /// <returns>Default(<typeparamref name="T"/>) pokud je seznam prázdný, 
    /// jinak data typu <typeparamref name="T"/>.</returns>
    public T OdeberPosledni()
    {
      if (JePrazdny)
        return default(T);

      PrvekObousmerehoSeznamu pom1 = posledni;
      if (aktualni == posledni)
      {
        if (aktualni == prvni)
          prvni = prvni.Predchudce;

        aktualni = posledni.Predchudce;
        posledni = aktualni;
      }
      else
        posledni = posledni.Predchudce;

      if (posledni != null)
        posledni.Naslednik = null;
      return pom1.data;
    }

    /// <summary>
    /// Odebere aktuální prvek seznamu.
    /// <remarks>Reference aktuálního prvku se přenese na následníka, nebo předchůdce, 
    /// pokud je akt. prvek posledním prvkem.</remarks>
    /// </summary>
    /// <returns>Default(<typeparamref name="T"/>) pokud je seznam prázdný, 
    /// jinak data typu <typeparamref name="T"/>.</returns>
    public T OdeberAktualni()
    {
      if (JePrazdny)
        return default(T);

      if (aktualni == prvni)
      {
        return OdeberPrvni();
      }

      if (aktualni == posledni)
      {
        return OdeberPosledni();
      }
      else
      {
        PrvekObousmerehoSeznamu pom1 = aktualni.Predchudce;
        PrvekObousmerehoSeznamu pom2 = aktualni.Naslednik;
        T pomData = aktualni.data;
        aktualni = pom2;
        pom2.Predchudce = pom1;
        pom1.Naslednik = pom2;
        return pomData;
      }
    }

    /// <summary>
    /// Odebere následníka aktuálního prvku.
    /// </summary>
    /// <returns>Default(<typeparamref name="T"/>) pokud je seznam prázdný, 
    /// nebo neexistuje následník aktuálního prvku,
    /// jinak data typu <typeparamref name="T"/>.</returns>
    public T OdeberNaslednika()
    {
      if (aktualni == null)
        return default(T);
      else
        if (aktualni.Naslednik == null)
          return default(T);

      PrvekObousmerehoSeznamu pom = aktualni.Naslednik;
      if (pom == posledni)
        posledni = aktualni;
      T pomData = pom.data;
      pom = pom.Naslednik;
      if (pom != null)
        pom.Predchudce = aktualni;
      aktualni.Naslednik = pom;
      return pomData;
    }

    /// <summary>
    /// Odebere předchůdce aktuálního prvku.
    /// </summary>
    /// <returns>Default(<typeparamref name="T"/>) pokud je seznam prázdný, 
    /// nebo neexistuje předchůdce aktuálního prvku,
    /// jinak data typu <typeparamref name="T"/>.</returns>
    public T OdeberPredchudce()
    {
      if (aktualni == null)
        return default(T);
      else
        if (aktualni.Predchudce == null)
          return default(T);

      PrvekObousmerehoSeznamu pom = aktualni.Predchudce;
      if (pom == prvni)
        prvni = aktualni;
      T pomData = pom.data;
      pom = pom.Predchudce;
      if (pom != null)
        pom.Naslednik = aktualni;
      aktualni.Predchudce = pom;
      return pomData;
    }

    /// <summary>
    /// Zpřístupní první prvek seznamu. Pokud existuje první prvek, reference akt. prvku 
    /// se přenese na první prvek.
    /// </summary>
    /// <returns>Default(<typeparamref name="T"/>) pokud je seznam prázdný, 
    /// jinak data typu <typeparamref name="T"/>.</returns>
    public T ZpristupniPrvni()
    {
      if (JePrazdny)
        return default(T);
      aktualni = prvni;
      return prvni.data;
    }

    /// <summary>
    /// Zpřístupní poslední prvek seznamu. Pokud existuje poslední prvek, reference akt. prvku 
    /// se přenese na poslední prvek.
    /// </summary>
    /// <returns>Default(<typeparamref name="T"/>) pokud je seznam prázdný, 
    /// jinak data typu <typeparamref name="T"/>.</returns>
    public T ZpristupniPosledni()
    {
      if (JePrazdny)
        return default(T);
      aktualni = posledni;
      return posledni.data;
    }

    /// <summary>
    /// Zpřístupní aktuální prvek seznamu.
    /// </summary>
    /// <returns>Default(<typeparamref name="T"/>) pokud je seznam prázdný, 
    /// jinak data typu <typeparamref name="T"/>.</returns>
    public T ZpristupniAktualni()
    {
      if (JePrazdny)
        return default(T);
      return aktualni.data;
    }

    /// <summary>
    /// Zpřístupní předchůdce aktuálního prvku. Pokud existuje předchůdce akt. prvku, 
    /// reference akt. prvku se přenese na předchůdce.
    /// </summary>
    /// <returns>Default(<typeparamref name="T"/>) pokud je seznam prázdný, 
    /// nebo neexistuje předchůdce aktuálního prvku,
    /// jinak data typu <typeparamref name="T"/>.</returns>
    public T ZpristupniPredchudce()
    {
      if (JePrazdny)
        return default(T);
      else
        if (aktualni.Predchudce == null)
          return default(T);

      aktualni = aktualni.Predchudce;
      return aktualni.data;
    }

    /// <summary>
    /// Zpřístupní následníka aktuálního prvku. Pokud existuje následník akt. prvku, 
    /// reference akt. prvku se přenese na následníka.
    /// </summary>
    /// <returns>Default(<typeparamref name="T"/>) pokud je seznam prázdný, 
    /// nebo neexistuje následník aktuálního prvku,
    /// jinak data typu <typeparamref name="T"/>.</returns>
    public T ZpristupniNaslednika()
    {
      if (JePrazdny)
        return default(T);
      else
        if (aktualni.Naslednik == null)
          return default(T);

      aktualni = aktualni.Naslednik;
      return aktualni.data;
    }

    /// <summary>
    /// Vytvoří enumerátor.
    /// </summary>
    /// <returns>
    /// Enumerátor seznamu.</returns>
    public IEnumerator<T> GetEnumerator()
    {
      if (!JePrazdny)
      {
        PrvekObousmerehoSeznamu prvek = prvni;

        while (prvek != null)
        {
          yield return prvek.data;
          prvek = prvek.Naslednik;
        }
      }
    }

    System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
    {
      return GetEnumerator();
    }
    #endregion


    #region Konstruktor
    /// <summary>
    /// Bezparametrický konstruktor.
    /// </summary>
    public AdsObousmernySeznam() { }

    /// <summary>
    /// Parametrický konstruktor. Vloží prvek s danými daty do seznamu.
    /// </summary>
    /// <param name="data"></param>
    public AdsObousmernySeznam(T data)
    {
      VlozPrvni(data);
    }
    #endregion
  }
}