﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using MG.DatoveStruktury.LinearniMnoziny;


namespace MG.DatoveStruktury.Tabulky
{
  /// <summary>
  /// Abstraktní datová struktura Binární vyhledávací strom. Strom je nevyvážený, 
  /// průměrná složitost operací vyhledávání/vkládání/odstraňování
  /// prvků je průměrně log(n) o základu 2, kde n je počet prvků ve stromě. 
  /// <remarks>Ve speciálních případech může strom zdegenerovat.</remarks>
  /// </summary>
  /// <typeparam name="TData">Referenční datový typ, které májí obsahovat prvky stromu.</typeparam>
  /// <typeparam name="TKlic">Referenční datový typ, který je klíčem prvku. 
  /// Typ klíče musí implementovat rozhraní <see cref="IComparable"/>.</typeparam>
  [Serializable]
  public class AdsBinarniVyhledavaciStrom<TData, TKlic> : IEnumerable<IPrvekTabulky<TData, TKlic>> where TKlic : IComparable<TKlic>
  {
    #region Interní datové typy
    [Serializable]
    class PrvekBVS : IPrvekTabulky<TData, TKlic>
    {
      #region Datové složky
      public PrvekBVS LevySyn { get; set; }

      public PrvekBVS PravySyn { get; set; }

      public PrvekBVS Otec { get; set; }

      public TData Data { get; set; }

      public TKlic Klic { get; set; }
      #endregion


      #region Vlastnosti
      /// <summary>
      /// True je-li prvek stromu listem, false jinak.
      /// </summary>
      public bool JeListem
      {
        get { return (LevySyn == null && PravySyn == null); }
      }
      #endregion


      #region Konstruktor
      internal PrvekBVS(TData data, TKlic klic)
      {
        LevySyn = null;
        PravySyn = null;
        Otec = null;
        this.Data = data;
        this.Klic = klic;
      }
      #endregion
    }

    [Serializable]
    class IteratorInOrderBVS
    {
      #region Datové složky
      private AdsZasobnik<PrvekBVS> zasobnik;
      #endregion

      #region Metody
      internal bool Dalsi(out PrvekBVS prvek)
      {
        if (zasobnik.JePrazdny)
        {
          prvek = null;
          return false;
        }

        PrvekBVS pomPrvek = zasobnik.odeberPrvek();
        PrvekBVS vysledek = pomPrvek;

        if (pomPrvek.PravySyn != null)
        {
          pomPrvek = pomPrvek.PravySyn;
          zasobnik.VlozPrvek(pomPrvek);

          while (pomPrvek.LevySyn != null)
          {
            pomPrvek = pomPrvek.LevySyn;
            zasobnik.VlozPrvek(pomPrvek);
          }
        }

        prvek = vysledek;
        return true;
      }
      #endregion


      #region Konstruktor
      internal IteratorInOrderBVS(PrvekBVS koren)
      {
        zasobnik = new AdsZasobnik<PrvekBVS>();
        if (koren != null)
        {
          zasobnik.VlozPrvek(koren);

          PrvekBVS pomPrvek = koren;
          while (pomPrvek.LevySyn != null)
          {
            pomPrvek = pomPrvek.LevySyn;
            zasobnik.VlozPrvek(pomPrvek);
          }
        }
      }
      #endregion
    }
    #endregion


    #region Datové složky
    private PrvekBVS koren;

    private int mohutnost = 0;

    /// <summary>
    /// Z důvodu co možná nejvetší vyváženosti stromu se odstranění prvku jednou provádí 
    /// náhradou nejbližšího vyššího prvku (z hlediska klíče),
    /// a jednou náhradou nejbližšího nižšího prvku.
    /// </summary>
    private bool nejblizsiNizsiHodnota = true;
    #endregion


    #region Vlastnosti
    /// <summary>
    /// Počet prvků ve struktuře.
    /// </summary>
    public int Mohutnost
    {
      get { return mohutnost; }
    }

    /// <summary>
    /// <c>True</c> je-li strom prázdný (obsahuje 0 prvků), <c>false</c> jinak.
    /// </summary>
    public bool JePrazdny
    {
      get { return koren == null; }
    }
    #endregion


    #region Metody
    private void Traverzuj(TKlic klic, out PrvekBVS aktualniPrvek)
    {
      bool konec = false;
      aktualniPrvek = koren;


      while (!konec)
      {
        if (aktualniPrvek.LevySyn != null && (klic.CompareTo(aktualniPrvek.Klic) < 0))
        {
          aktualniPrvek = aktualniPrvek.LevySyn;
          continue;
        }
        else
          if (aktualniPrvek.PravySyn != null && (klic.CompareTo(aktualniPrvek.Klic) > 0))
          {
            aktualniPrvek = aktualniPrvek.PravySyn;
            continue;
          }
          else
            konec = true;
      }
    }

    private void OdeberList(PrvekBVS list)
    {
      if (ReferenceEquals(list, list.Otec.LevySyn))
      {

        list.Otec.LevySyn = null;
      }
      else
      {
        list.Otec.PravySyn = null;
      }
    }

    private void odeberKoren(PrvekBVS staryKoren)
    {
      if (staryKoren.LevySyn != null)
      {
        koren = koren.LevySyn;
        koren.Otec = null;
      }
      else
      {
        koren = koren.PravySyn;
        koren.Otec = null;
      }
    }

    private PrvekBVS NajdiNejblizsiNizsiHodnotuKlice(PrvekBVS prvek)
    {
      PrvekBVS vysledek = prvek;
      bool konec = false;

      if (vysledek.LevySyn != null)
      {
        vysledek = vysledek.LevySyn;
        while (!konec)
        {
          if (vysledek.PravySyn != null)
          {
            vysledek = vysledek.PravySyn;
          }
          else
            konec = true;
        }
      }

      return vysledek;
    }

    private PrvekBVS NajdiNejblizsiVyssiHodnotuKlice(PrvekBVS prvek)
    {
      PrvekBVS vysledek = prvek;
      bool konec = false;

      if (vysledek.PravySyn != null)
      {
        vysledek = vysledek.PravySyn;
        while (!konec)
        {
          if (vysledek.LevySyn != null)
          {
            vysledek = vysledek.LevySyn;
          }
          else
            konec = true;
        }
      }

      return vysledek;
    }

    private void PrepojSynaKOtci(PrvekBVS prvek)
    {
      if (prvek.LevySyn != null)
      {
        if (ReferenceEquals(prvek.Otec.LevySyn, prvek))
        {
          prvek.Otec.LevySyn = prvek.LevySyn;
          prvek.LevySyn.Otec = prvek.Otec;
        }
        else
        {
          prvek.Otec.PravySyn = prvek.LevySyn;
          prvek.LevySyn.Otec = prvek.Otec;
        }
      }
      else
      {
        if (ReferenceEquals(prvek.Otec.LevySyn, prvek))
        {
          prvek.Otec.LevySyn = prvek.PravySyn;
          prvek.PravySyn.Otec = prvek.Otec;
        }
        else
        {
          prvek.Otec.PravySyn = prvek.PravySyn;
          prvek.PravySyn.Otec = prvek.Otec;
        }
      }
    }

    private void VymenDataKlic(PrvekBVS prvek1, PrvekBVS prvek2)
    {
      TData pomPrvekData = prvek1.Data;
      TKlic pomPrvekKlic = prvek1.Klic;
      prvek1.Data = prvek2.Data;
      prvek1.Klic = prvek2.Klic;
      prvek2.Data = pomPrvekData;
      prvek2.Klic = pomPrvekKlic;
    }

    /// <summary>
    /// Vloží prvek do stromu.
    /// </summary>
    /// <param name="data">Data, která má prvek obsahovat.</param>
    /// <param name="klic">Unikátní klíč prvku.</param>
    /// <exception cref="System.ArgumentException">Metoda vyvolá výjimku v připadě vkládání duplicitního klíče.</exception>
    public void VlozPrvek(TData data, TKlic klic)
    {
      if (JePrazdny)
      {
        koren = new PrvekBVS(data, klic);
        mohutnost++;
        return;
      }

      PrvekBVS aktualniPrvek = null;
      Traverzuj(klic, out aktualniPrvek);

      if (aktualniPrvek.Klic.CompareTo(klic) == 0)
        throw new ArgumentException("Prvek s daným klíčem ve stromě již existuje.", "klic");


      if (klic.CompareTo(aktualniPrvek.Klic) < 0)
      {
        aktualniPrvek.LevySyn = new PrvekBVS(data, klic);
        aktualniPrvek.LevySyn.Otec = aktualniPrvek;
        mohutnost++;
      }
      else
        if (klic.CompareTo(aktualniPrvek.Klic) > 0)
        {
          aktualniPrvek.PravySyn = new PrvekBVS(data, klic);
          aktualniPrvek.PravySyn.Otec = aktualniPrvek;
          mohutnost++;
        }
    }

    /// <summary>
    /// Nalezne data prvku ve stromě dle daného klíče.
    /// </summary>
    /// <param name="klic">Unikátní klíč prvku.</param>
    /// <returns>Data typu <typeparamref name="TData"/>
    /// nebo Default(<typeparamref name="TData"/>) v případě, že prvek s daným klíčem ve stromu neexistuje.</returns>
    public TData NajdiData(TKlic klic)
    {
      IPrvekTabulky<TData, TKlic> prvek = NajdiPrvek(klic);

      if (prvek != default(IPrvekTabulky<TData, TKlic>))
      {
        return prvek.Data;
      }
      else
      {
        return default(TData);
      }
    }

    /// <summary>
    /// Nalezne prvek ve stromě dle daného klíče.
    /// </summary>
    /// <param name="klic">Unikátní klíč prvku.</param>
    /// <returns>Data typu <see cref="IPrvekTabulky{TData, TKlic}"/>
    /// nebo Default(<see cref="IPrvekTabulky{TData, TKlic}"/>) v případě, že prvek s daným klíčem ve stromu neexistuje.</returns>
    public IPrvekTabulky<TData, TKlic> NajdiPrvek(TKlic klic)
    {
      if (JePrazdny)
        return default(IPrvekTabulky<TData, TKlic>);

      PrvekBVS aktualniPrvek = null;
      Traverzuj(klic, out aktualniPrvek);
      if (aktualniPrvek.Klic.CompareTo(klic) == 0)
        return aktualniPrvek as IPrvekTabulky<TData, TKlic>;

      if (aktualniPrvek.LevySyn != null)
      {
        if (aktualniPrvek.LevySyn.Klic.CompareTo(klic) == 0)
        {
          aktualniPrvek = aktualniPrvek.LevySyn;
          return aktualniPrvek as IPrvekTabulky<TData, TKlic>;
        }
      }

      if (aktualniPrvek.PravySyn != null)
      {
        if (aktualniPrvek.PravySyn.Klic.CompareTo(klic) == 0)
        {
          aktualniPrvek = aktualniPrvek.PravySyn;
          return aktualniPrvek as IPrvekTabulky<TData, TKlic>;
        }
      }

      return default(IPrvekTabulky<TData, TKlic>);
    }

    /// <summary>
    /// Najde a popřípadě odstraní prvek podle daného klíče.
    /// </summary>
    /// <param name="klic">Unikátní klíč prvku.</param>
    /// <returns>Data typu <typeparamref name="TData"/> 
    /// nebo Default(<typeparamref name="TData"/>) v případě, že prvek dle klíče se ve stromu nenachází.</returns>
    public TData OdstranPrvek(TKlic klic)
    {
      if (JePrazdny)
        return default(TData);

      PrvekBVS aktualniPrvek = NajdiPrvek(klic) as PrvekBVS;
      if (aktualniPrvek == null || aktualniPrvek.Klic.CompareTo(klic) != 0)
        return default(TData);

      if (koren.LevySyn == null && koren.PravySyn == null) // je tam jen kořen
      {
        TData vysledekData = aktualniPrvek.Data;
        aktualniPrvek = null;
        koren = null;
        mohutnost--;
        return vysledekData;
      }
      else
        if (aktualniPrvek.JeListem) // odebiram list
        {
          OdeberList(aktualniPrvek);
          mohutnost--;
          TData d = aktualniPrvek.Data;
          aktualniPrvek = koren;
          return d;
        }
        else
          if ((aktualniPrvek.LevySyn != null || aktualniPrvek.PravySyn != null) && 
            !(aktualniPrvek.LevySyn != null && aktualniPrvek.PravySyn != null)) // prvek má pouze 1 syna
          {
            if (aktualniPrvek.Otec != null)
            {
              PrepojSynaKOtci(aktualniPrvek);
              mohutnost--;
              TData d = aktualniPrvek.Data;
              aktualniPrvek = koren;
              return d;
            }
            else
            {
              TData d = koren.Data;
              odeberKoren(aktualniPrvek); // odebírá se kořen s 1 synem
              mohutnost--;
              return d;
            }
          }
          else // odebírám prvek se 2 syny
          {
            PrvekBVS nejblizsiPrvek = null;
            if (nejblizsiNizsiHodnota)
            {
              nejblizsiPrvek = NajdiNejblizsiNizsiHodnotuKlice(aktualniPrvek);
              nejblizsiNizsiHodnota = false;
            }
            else
            {
              nejblizsiPrvek = NajdiNejblizsiVyssiHodnotuKlice(aktualniPrvek);
              nejblizsiNizsiHodnota = true;
            }

            if (!ReferenceEquals(aktualniPrvek, nejblizsiPrvek))
            {
              VymenDataKlic(nejblizsiPrvek, aktualniPrvek);
            }

            if (nejblizsiPrvek.JeListem)
            {
              TData d = nejblizsiPrvek.Data;
              OdeberList(nejblizsiPrvek);
              mohutnost--;
              aktualniPrvek = koren;
              return d;
            }
            else
            {
              TData d = nejblizsiPrvek.Data;
              PrepojSynaKOtci(nejblizsiPrvek);
              mohutnost--;
              aktualniPrvek = koren;
              return d;
            }
          }
    }

    /// <summary>
    /// Vytvoří enumerátor. Prohlídka InOrder.
    /// </summary>
    /// <returns>Enumerátor stromu.</returns>
    public IEnumerator<IPrvekTabulky<TData, TKlic>> GetEnumerator()
    {
      PrvekBVS prvek = null;
      IteratorInOrderBVS iteratorBVS = new IteratorInOrderBVS(koren);
      while (iteratorBVS.Dalsi(out prvek))
      {
        yield return prvek as IPrvekTabulky<TData, TKlic>;
      }
      iteratorBVS = null;
    }

    System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
    {
      return GetEnumerator();
    }
    #endregion


    #region Konstruktor
    /// <summary>
    /// Bezparametrický konstruktor.
    /// </summary>
    public AdsBinarniVyhledavaciStrom() { }

    /// <summary>
    /// Parametrický konstruktor. Vloží kořen do stromu dle parametru.
    /// </summary>
    /// <param name="data">Data, která ponese kořen.</param>
    /// <param name="klic">Unikátní klíč.</param>
    public AdsBinarniVyhledavaciStrom(TData data, TKlic klic)
    {
      koren = new PrvekBVS(data, klic);
    }
    #endregion
  }
}