﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace MG.DatoveStruktury.Tabulky
{
  /// <summary>
  /// Abstraktní datová struktura Tabulka implementovaná pomocí jednosměrného lineárního necyklického seznamu. Tabulka je utříděná,
  /// složitost všech operací je maximálně O(n), kde n je počet prvků tabulky.
  /// </summary>
  /// <typeparam name="TData">Referenční datový typ, který obsahují prvky Tabulky.</typeparam>
  /// <typeparam name="TKlic">Referenční datový typ, který je klíčem prvku. 
  /// Typ klíče musí implementovat rozhraní <see cref="IComparable"/>.</typeparam>
  [Serializable]
  public class AdsTabulkaSeznam<TData, TKlic> : IEnumerable<IPrvekTabulky<TData, TKlic>>
    where TKlic : IComparable<TKlic>
  {
    #region Interní datové typy
    [Serializable]
    class PrvekTabulky : IPrvekTabulky<TData, TKlic>
    {
      #region Datové složky
      public PrvekTabulky Naslednik { get; set; }

      public TData Data { get; set; }

      public TKlic Klic { get; set; }
      #endregion


      #region Konstruktor
      internal PrvekTabulky(TData data, TKlic klic)
      {
        this.Data = data;
        this.Klic = klic;
        this.Naslednik = null;
      }
      #endregion
    }
    #endregion


    #region Datové položky
    private PrvekTabulky prvni;

    private int mohutnost = 0;
    #endregion


    #region Vlastnosti
    /// <summary>
    /// <c>Ttrue</c> je-li tabulka prázdná, <c>False</c> jinak.
    /// </summary>
    public bool JePrazdna
    {
      get
      {
        return prvni == null;
      }
    }

    /// <summary>
    /// Počet prvků v tabulce.
    /// </summary>
    public int Mohutnost
    {
      get
      { return mohutnost; }
    }
    #endregion


    #region Metody
    /// <summary>
    /// Vloží prvek s danými daty a klíčem do tabulky.
    /// </summary>
    /// <param name="data">Data typu <c><typeparamref name="TData"/></c>, která má prvek obsahovat.</param>
    /// <param name="klic">Unikátní klíč prvku.</param>
    /// <exception cref="System.ArgumentException">Metoda vyvolá výjimku v připadě vkládání duplicitního klíče.</exception>
    public void VlozPrvek(TData data, TKlic klic)
    {
      PrvekTabulky vkladany = new PrvekTabulky(data, klic), aktualni = null, preaktualni = null;

      if (JePrazdna)
      {
        prvni = vkladany;
        mohutnost++;
        return;
      }

      aktualni = prvni;
      while (vkladany.Klic.CompareTo(aktualni.Klic) > 0)
      {
        preaktualni = aktualni;
        aktualni = aktualni.Naslednik;
        if (aktualni == null)
        {
          preaktualni.Naslednik = vkladany;
          mohutnost++;
          return;
        }
      }

      if (vkladany.Klic.CompareTo(aktualni.Klic) == 0)
        throw new ArgumentException("Prvek s daným klíčem v tabulce již existuje.", "klic");

      if (preaktualni != null)
        preaktualni.Naslednik = vkladany;
      else
        prvni = vkladany;
      vkladany.Naslednik = aktualni;
      mohutnost++;
    }

    /// <summary>
    /// Odebere prvek z tabulky podle zadaného klíče.
    /// </summary>
    /// <param name="klic">Klíč hledaného prvku.</param>
    /// <returns>Default(<typeparamref name="TData"/>), pokud prvek neexistuje, 
    /// jinak data typu <typeparamref name="TData"/>.</returns>
    public TData OdeberPrvek(TKlic klic)
    {
      if (JePrazdna)
        return default(TData);

      PrvekTabulky aktualni = prvni.Naslednik;
      PrvekTabulky preaktualni = prvni;

      if (preaktualni.Klic.CompareTo(klic) == 0)
      {
        prvni = aktualni;
        mohutnost--;
        return preaktualni.Data;
      }

      while (aktualni != null)
      {
        if (aktualni.Klic.CompareTo(klic) == 0)
        {
          TData pomData = aktualni.Data;
          aktualni = aktualni.Naslednik;
          preaktualni.Naslednik = aktualni;
          mohutnost--;
          return pomData;
        }

        if (aktualni.Klic.CompareTo(klic) > 0)
          break;

        preaktualni = aktualni;
        aktualni = aktualni.Naslednik;
      }

      return default(TData);
    }

    /// <summary>
    /// Zpřístupní datovou část prvku z tabulky dle zadaného klíče.
    /// </summary>
    /// <param name="klic">Klíč hledaného prvku.</param>
    /// <returns>Default(<typeparamref name="TData"/>), pokud prvek neexistuje, 
    /// jinak data typu <typeparamref name="TData"/>.</returns>
    public TData NajdiData(TKlic klic)
    {
      IPrvekTabulky<TData, TKlic> prvek = NajdiPrvek(klic);

      if (prvek != default(IPrvekTabulky<TData, TKlic>))
      {
        return prvek.Data;
      }
      else
      {
        return default(TData);
      }
    }

    /// <summary>
    /// Zpřístupní prvek z tabulky dle zadaného klíče.
    /// </summary>
    /// <param name="klic">Klíč hledaného prvku.</param>
    /// <returns>Default(<see cref="IPrvekTabulky{TData, TKlic}"/>), pokud prvek neexistuje, 
    /// jinak data typu <see cref="IPrvekTabulky{TData, TKlic}"/>.</returns>
    public IPrvekTabulky<TData, TKlic> NajdiPrvek(TKlic klic)
    {
      if (JePrazdna)
        return default(IPrvekTabulky<TData, TKlic>);

      PrvekTabulky aktualni = prvni;

      while (aktualni != null)
      {
        if (aktualni.Klic.CompareTo(klic) == 0)
          return aktualni as IPrvekTabulky<TData, TKlic>;
        aktualni = aktualni.Naslednik;

        if (aktualni != null && aktualni.Klic.CompareTo(klic) > 0)
          break;
      }

      return default(IPrvekTabulky<TData, TKlic>);
    }

    /// <summary>
    /// Vytvoří enumeríátor.
    /// </summary>
    /// <returns>Enumerátor tabulky.</returns>
    public IEnumerator<IPrvekTabulky<TData, TKlic>> GetEnumerator()
    {
      if (!JePrazdna)
      {
        PrvekTabulky prvek = prvni;

        while (prvek != null)
        {
          yield return (prvek as IPrvekTabulky<TData, TKlic>);
          prvek = prvek.Naslednik;
        }
      }
    }

    System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
    {
      return GetEnumerator();
    }
    #endregion


    #region Konstruktor
    /// <summary>
    /// Bezparametrický konstruktor.
    /// </summary>
    public AdsTabulkaSeznam() { }

    /// <summary>
    /// Parametrický konstruktor. Vloží prvek s danými daty a klíčem do tabulky.
    /// </summary>
    /// <param name="data">Data typu <c><typeparamref name="TData"/></c>, která má prvek obsahovat.</param>
    /// <param name="klic">Unikátní klíč prvku.</param>
    public AdsTabulkaSeznam(TData data, TKlic klic)
    {
      VlozPrvek(data, klic);
    }
    #endregion
  }
}