﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace MG.Matematika.Geometrie2D
{
  /// <summary>
  /// Systém uchovávání 2D souřadného systému. obsahuje potřebné datové složky a metody.
  /// </summary>
  public class SouradnySystem2D
  {
    #region Datové složky
    private int rozpetiPX;
    private int rozpetiPY;
    private double realMinX;
    private double realMaxX;
    private double realMinY;
    private double realMaxY;
    /// <summary>
    /// Pomocné tatové složky pro zjištění poměru mezi souřadnicemi reálného souřadného systému
    /// a pixelovým souřadným systémem.
    /// </summary>
    private double dx, dy;
    #endregion


    #region Vlastnosti
    /// <summary>
    /// Počet jednotek pixelového souřadného systému ve směru x.
    /// <remarks>Validní jsou pouze kladné hodnoty, záporné hodnoty se automaticky převádějí na kladné</remarks>
    /// </summary>
    public int RozpetiPX
    {
      get { return rozpetiPX; }
      set { rozpetiPX = Math.Abs(value); ZjistiKoeficienty(); }
    }

    /// <summary>
    /// Počet jednotek pixelového souřadného systému ve směru y.
    /// <remarks>Validní jsou pouze kladné hodnoty, záporné hodnoty se automaticky převádějí na kladné</remarks>
    /// </summary>
    public int RozpetiPY
    {
      get { return rozpetiPY; }
      set { rozpetiPY = Math.Abs(value); ZjistiKoeficienty(); }
    }

    /// <summary>
    /// Minimální X-ová hodnota reálného souřadného systému.
    /// </summary>
    public double RealMinX { get { return realMinX; } set { realMinX = value; ZjistiKoeficienty(); } }

    /// <summary>
    /// Maximální X-ová hodnota reálného souřadného systému.
    /// </summary>
    public double RealMaxX { get { return realMaxX; } set { realMaxX = value; ZjistiKoeficienty(); } }

    /// <summary>
    /// Minimální Y-ová hodnota reálného souřadného systému.
    /// </summary>
    public double RealMinY { get { return realMinY; } set { realMinY = value; ZjistiKoeficienty(); } }

    /// <summary>
    /// Maximální Y-ová hodnota reálného souřadného systému.
    /// </summary>
    public double RealMaxY { get { return realMaxY; } set { realMaxY = value; ZjistiKoeficienty(); } }
    #endregion


    #region Metody
    /// <summary>
    /// Nastaví pomocné atributy dx a dy podle aktuálního poměru reálné ss a pixelové ss.
    /// </summary>
    private void ZjistiKoeficienty()
    {
      dx = rozpetiPX / (RealMaxX - RealMinX);
      dy = rozpetiPY / (RealMaxY - RealMinY);
    }

    /// <summary>
    /// Převede souřadnice z reálné ss do pixelové ss.
    /// </summary>
    /// <param name="rBod">Bod se souřadnicemi reálné souřadné soustavy.</param>
    /// <returns><paramref name="rBod"/> s převedenými souřadnicemi do pixelové ss.</returns>
    public Bod<int> R2P(Bod<double> rBod)
    {
      Bod<int> vysledek = new Bod<int>();
      vysledek.X = (int)(Math.Round((rBod.X - RealMinX) * dx));
      vysledek.Y = (int)(Math.Round((-rBod.Y + RealMaxY) * dy));
      return vysledek;
    }

    /// <summary>
    /// Převede X-sovou reálnou souřadnici do X-ové pixelové souřadnice.
    /// </summary>
    /// <param name="rSouradnice">X-ová reálná souřadnice</param>
    /// <returns>X-ová pixelová souřadnice.</returns>
    public int R2PX(double rSouradnice)
    {
      return (int)(Math.Round((rSouradnice - RealMinX) * dx));
    }

    /// <summary>
    /// Převede Y-sovou reálnou souřadnici do Y-ové pixelové souřadnice.
    /// </summary>
    /// <param name="rSouradnice">Y-ová reálná souřadnice</param>
    /// <returns>Y-ová pixelová souřadnice.</returns>
    public int R2PY(double rSouradnice)
    {
      return (int)(Math.Round((-rSouradnice + RealMaxY) * dy));
    }

    /// <summary>
    /// Převede souřadnice z pixelové ss do reálné ss.
    /// </summary>
    /// <param name="pBod">Bod se souřadnicemi pixelové souřadné soustavy.</param>
    /// <returns>Bod s převedenými souřadnicemi do reálné ss.</returns>
    public Bod<double> P2R(Bod<int> pBod)
    {
      Bod<double> vysledek = new Bod<double>();
      vysledek.X = RealMinX + pBod.X / dx;
      vysledek.Y = RealMaxY - pBod.Y / dy;
      return vysledek;
    }

    /// <summary>
    /// Převede X-sovou pixelovou souřadnici do X-ové reálné souřadnice.
    /// </summary>
    /// <param name="pSouradnice">X-ová pixelová souřadnice</param>
    /// <returns>X-ová reálná souřadnice.</returns>
    public double P2RX(int pSouradnice)
    {
      return RealMinX + pSouradnice / dx;
    }

    /// <summary>
    /// Převede Y-sovou pixelovou souřadnici do Y-ové reálné souřadnice.
    /// </summary>
    /// <param name="pSouradnice">Y-ová pixelová souřadnice</param>
    /// <returns>Y-ová reálná souřadnice.</returns>
    public double P2RY(int pSouradnice)
    {
      return RealMaxY - pSouradnice / dy;
    }

    /// <summary>
    /// Posune kreslící plátno. 
    /// <remarks>Po volání funkce by se měla volat metoda Invalidate 
    /// dané komponenty (pokud používá ss ke grafickému vykreslování). </remarks>
    /// </summary>
    /// <param name="posunX">Posun ve vodorovném směru.</param>
    /// <param name="posunY">Posun ve svislém směru.</param>
    public void NavigacePosun(double posunX, double posunY)
    {
      RealMinX += posunX;
      RealMaxX += posunX;
      RealMinY += posunY;
      RealMaxY += posunY;
    }

    /// <summary>
    /// Posune kreslící plátno rovným dílem jako je posun pixelový
    /// <remarks>Po volání funkce by se měla volat metoda Invalidate 
    /// dané komponenty (pokud používá ss ke grafickému vykreslování). </remarks>
    /// </summary>
    /// <param name="posunX">Posun ve vodorovném směru.</param>
    /// <param name="posunY">Posun ve svislém směru.</param>
    public void NavigacePosunPX(int posunX, int posunY)
    {
      RealMinX += posunX / (dx);
      RealMaxX += posunX / (dx);
      RealMinY += posunY / (dy);
      RealMaxY += posunY / (dy);
    }

    /// <summary>
    /// Provede změnu "přiblížení" souřadného systému změnou rozpětí realného souřadného systému.
    /// <remarks>Po volání funkce by se měla volat metoda Invalidate 
    /// dané komponenty (pokud používá ss ke grafickému vykreslování). </remarks>
    /// </summary>
    /// <param name="zlomekZmenyRozpeti">Dělitel současného rozpětí, který se přičte s současnému rozpetí. 
    /// Rozpětí se mění rovným dílem jak pro osu x, tak pro osu y.</param>
    public void NavigaceZoom(double zlomekZmenyRozpeti)
    {
        RealMinX += RealMinX / zlomekZmenyRozpeti;
        RealMaxX += RealMaxX / zlomekZmenyRozpeti;
        RealMinY += RealMinY / zlomekZmenyRozpeti;
        RealMaxY += RealMaxY / zlomekZmenyRozpeti;
      ZjistiKoeficienty();
    }
    #endregion
  }
}
