﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using MG.DatoveStruktury.LinearniMnoziny;
using MG.DatoveStruktury.Tabulky;
using MG.DatoveStruktury.Grafy.SmisenaReprezentace;

namespace MG.OperacniVyzkum.TeorieGrafu
{
  /// <summary>
  /// Třída obsahuje metody pro výpočty cest na grafech.
  /// </summary>
  /// <typeparam name="TVrchol">Datový typ vrcholů.</typeparam>
  /// <typeparam name="THrana">Datový typ hran.</typeparam>
  /// <typeparam name="TKlicVrcholu">Datový typ pro Klíč vrcholu.</typeparam>
  /// <typeparam name="TKlicHrany">Datový typ pro Klíč hrany.</typeparam>
  /// <typeparam name="TOhodnoceniHrany">Datový typ pro ohodnocení hrany.</typeparam>
  public static class CestyNaGrafech<TVrchol, THrana, TKlicVrcholu, TKlicHrany, TOhodnoceniHrany>
    where TVrchol : Vrchol<TKlicVrcholu, TKlicHrany, TOhodnoceniHrany>
    where THrana : Hrana<TKlicHrany, TKlicVrcholu, TOhodnoceniHrany>
    where TKlicVrcholu : IComparable<TKlicVrcholu>
    where TKlicHrany : IComparable<TKlicHrany>
    where TOhodnoceniHrany : IComparable<TOhodnoceniHrany>
  {
    #region Interní datové typy
    /// <summary>
    /// Potřebné informace o vrcholu pro výpočet nejkratší cesty.
    /// </summary>
    [Serializable]
    internal class InformaceVrcholNejkratsiCesta
    {
      #region Vlastnosti
      /// <summary>
      /// Ohodnocení vrcholu.
      /// </summary>
      public double Ohodnceni { get; set; }
      #endregion


      #region Konstruktor
      /// <summary>
      /// Parametrický konstruktor.
      /// </summary>
      /// <param name="ohodnoceni">Ohodnocení vrcholu.</param>
      internal InformaceVrcholNejkratsiCesta(double ohodnoceni)
      {
        Ohodnceni = ohodnoceni;
      }
      #endregion
    }

    /// <summary>
    /// Potřebné informace o hraně pro výpočet nejkratší cesty.
    /// </summary>
    [Serializable]
    internal class InformaceHranaNejkratsiCesta
    {
      #region Vlastnosti
      /// <summary>
      /// Ohodnocení hrany.
      /// </summary>
      public double Ohodnceni { get; set; }
      #endregion


      #region Konstruktor
      /// <summary>
      /// Parametrický konstruktor.
      /// </summary>
      /// <param name="ohodnoceni">Ohodnocení hrany.</param>
      internal InformaceHranaNejkratsiCesta(double ohodnoceni)
      {
        Ohodnceni = ohodnoceni;
      }
      #endregion
    }

    internal class VrcholListComparer
      : IComparer<TVrchol>
    {
      public int Compare(TVrchol x, TVrchol y)
      {
        if ((x.Tag as InformaceVrcholNejkratsiCesta).Ohodnceni < (y.Tag as InformaceVrcholNejkratsiCesta).Ohodnceni)
          return -1;
        else if ((x.Tag as InformaceVrcholNejkratsiCesta).Ohodnceni > (y.Tag as InformaceVrcholNejkratsiCesta).Ohodnceni)
          return 1;
        else
          return 0;
      }
    }
    #endregion

    #region Metody
    #region Privátní
    /// <summary>
    /// Ohodnotí vrcholy Dijsktrovým algoritmem.
    /// </summary>
    /// <param name="vrchol">Zkoumaný vrchol.</param>
    /// <param name="pocatecniVrchol">Počátečný vrchol cesty.</param>
    /// <param name="koncovyVrchol">Koncový vrchol cesty.</param>
    /// <param name="list">List neprozkoumaných vrcholů.</param>
    private static void OhodnotVrcholyNejkratsiCesta(
      TVrchol vrchol,
      TVrchol pocatecniVrchol,
      TVrchol koncovyVrchol,
      List<TVrchol> list = null)
    {
      if (list == null)
        list = new List<TVrchol>();


      // Pro všechny incidenční hrany vrcholu
      foreach (THrana hrana in vrchol.SeznamSousedu)
      {
        // Zjistím kterým z incidenčních vrcholů je zkoumaný vrchol
        if (!ReferenceEquals(hrana.IncidencniVrchol1, vrchol))
        {// Pokud druhý vrchol není počátečním (ohodnocoval bych počáteční vrchol jinak)
          if (hrana.IncidencniVrchol1 != pocatecniVrchol)
          {
            double t_j = 
              Math.Round((hrana.Tag as InformaceHranaNejkratsiCesta).Ohodnceni +
              (hrana.IncidencniVrchol2.Tag as InformaceVrcholNejkratsiCesta).Ohodnceni, 6);
            // Pokud se najde kratší cesta k vrcholu než je současná
            if (t_j < (hrana.IncidencniVrchol1.Tag as InformaceVrcholNejkratsiCesta).Ohodnceni)
            {
              // Přehodnotím
              (hrana.IncidencniVrchol1.Tag as InformaceVrcholNejkratsiCesta).Ohodnceni = t_j;
              // Vložím vrchol do "prioritní fronty" k dalšímu hledání
              list.Add(hrana.IncidencniVrchol1 as TVrchol);
            }
          }
        }
        else
        {
          if (hrana.IncidencniVrchol2 != pocatecniVrchol)
          {
            double t_j = 
              Math.Round((hrana.Tag as InformaceHranaNejkratsiCesta).Ohodnceni +
              (hrana.IncidencniVrchol1.Tag as InformaceVrcholNejkratsiCesta).Ohodnceni, 6);

            if (t_j < (hrana.IncidencniVrchol2.Tag as InformaceVrcholNejkratsiCesta).Ohodnceni)
            {
              (hrana.IncidencniVrchol2.Tag as InformaceVrcholNejkratsiCesta).Ohodnceni = t_j;
              list.Add(hrana.IncidencniVrchol2 as TVrchol);
            }
          }
        }
      }

      list.Sort(0, list.Count, new VrcholListComparer());
      TVrchol vybranyVrchol = list[0];
      if (list[0] == koncovyVrchol)
        return;
      list.RemoveAt(0);
      OhodnotVrcholyNejkratsiCesta(vybranyVrchol, pocatecniVrchol, koncovyVrchol, list);
    }

    /// <summary>
    /// Metoda se pokusí najít zpětně nejkratší cestu od koncového vrcholu k počátečnímu.
    /// <exception cref="System.ArgumentException">Metoda vyvolá výjimku, pokud graf není obyčejný,
    /// nebo pokud graf není souvislý na hledané cestě.</exception>
    /// </summary>
    /// <param name="vrchol">Momentální vrchol cesty.</param>
    /// <param name="pocatecniVrchol">Počátečný vrchol cesty.</param>
    /// <returns></returns>
    private static AdsObousmernySeznam<IPrvekGrafu> NajdiCestuNejkratsiCesta(
      TVrchol vrchol,
      TVrchol pocatecniVrchol)
    {
      // Pokud je aktuální vrchol počátečný, dostalo se zpětně přes celou cestu
      if (vrchol == pocatecniVrchol)
      {
        return new AdsObousmernySeznam<IPrvekGrafu>(vrchol);
      }

      // Zde se bude stavět ekurzivně cesta
      AdsObousmernySeznam<IPrvekGrafu> vysledek = null;
      // Hrana, která se osvědčila jako část cesty
      Hrana<TKlicHrany, TKlicVrcholu, TOhodnoceniHrany> vybranaHrana = null;

      // Pro všechny incidenční hrany aktuálního vrcholu
      foreach (THrana hrana in vrchol.SeznamSousedu)
      {
        /*if (double.MaxValue == (hrana.IncidencniVrchol2.Tag as InformaceVrcholNejkratsiCesta).Ohodnceni ||
          double.MaxValue == (hrana.IncidencniVrchol1.Tag as InformaceVrcholNejkratsiCesta).Ohodnceni)
        {
          throw new ArgumentException("Graf není na dané cestě souvislý");
        }*/

        // Zjistím který z incidenčních vrcholů je ten aktuální
        if (ReferenceEquals(vrchol, hrana.IncidencniVrchol1))
        {
          double odecet = 
            Math.Round((vrchol.Tag as InformaceVrcholNejkratsiCesta).Ohodnceni - 
            (hrana.Tag as InformaceHranaNejkratsiCesta).Ohodnceni, 6);
          if ( // Pokud tudy vede cesta
            odecet == (hrana.IncidencniVrchol2.Tag as InformaceVrcholNejkratsiCesta).Ohodnceni)
          {
            // Rekurzivně hledám dál k počátku, výsledek popřípadě spojím
            vysledek = NajdiCestuNejkratsiCesta(hrana.IncidencniVrchol2 as TVrchol, pocatecniVrchol);
            vybranaHrana = hrana;
            break;
          }
        }
        else
        {
          double odecet = 
             Math.Round((vrchol.Tag as InformaceVrcholNejkratsiCesta).Ohodnceni -
             (hrana.Tag as InformaceHranaNejkratsiCesta).Ohodnceni, 6);
          if (odecet == (hrana.IncidencniVrchol1.Tag as InformaceVrcholNejkratsiCesta).Ohodnceni)
          {
            vysledek = NajdiCestuNejkratsiCesta(hrana.IncidencniVrchol1 as TVrchol, pocatecniVrchol);
            vybranaHrana = hrana;
            break;
          }
        }
      }

      // Pokud vnořené hledání vrátilo výsledek
      if (vysledek != null)
      {
        TVrchol iv =
          (vysledek.ZpristupniPosledni() as TVrchol);
        // Najdu který z incidenčních vrcholů hrany je ten v dosavadní cestě
        if (ReferenceEquals(iv, (vybranaHrana as Hrana<TKlicHrany, TKlicVrcholu, TOhodnoceniHrany>).IncidencniVrchol1))
        {
          // Vložím do cesty vrchol a hranu
          vysledek.VlozPosledni(vybranaHrana);
          vysledek.VlozPosledni((vybranaHrana as Hrana<TKlicHrany, TKlicVrcholu, TOhodnoceniHrany>).IncidencniVrchol2);
          // Předám o stupeň volání výše
          return vysledek;
        }

        if (ReferenceEquals(iv, (vybranaHrana as Hrana<TKlicHrany, TKlicVrcholu, TOhodnoceniHrany>).IncidencniVrchol2))
        {
          vysledek.VlozPosledni(vybranaHrana);
          vysledek.VlozPosledni((vybranaHrana as Hrana<TKlicHrany, TKlicVrcholu, TOhodnoceniHrany>).IncidencniVrchol1);

          return vysledek;
        }
        return null;
      }
      else
      {
        // Pokud se nenašlo pokračování cesty
        throw new ArgumentException("Graf není obyčejný nebo není na dané cestě souvislý");
      }
    }

    /// <summary>
    /// Metoda se pokusí najít nejkratší cestu dle zadaných parametrů.
    /// </summary>
    /// <param name="graf">Graf, na kterém se má hledat cesta.</param>
    /// <param name="pocatecniVrchol">Počátečný vrchol cesty.</param>
    /// <param name="koncovyVrchol">Koncový vrchol cesty.</param>
    /// <returns>Seznam prvků grafu tvořící cestu.</returns>
    private static AdsObousmernySeznam<IPrvekGrafu> _NejkratsiCesta(
      AdsGrafSmisenaReprezentace<TVrchol, THrana, TKlicVrcholu, TKlicHrany, TOhodnoceniHrany> graf,
      TVrchol pocatecniVrchol,
      TVrchol koncovyVrchol)
    {
      // Inicializace - značkování vrcholů--------------------------------------------------
      foreach (TVrchol vrchol in graf)
      {
        if (ReferenceEquals(vrchol, pocatecniVrchol))
        {
          vrchol.Tag = new InformaceVrcholNejkratsiCesta(0);
        }
        else
        {
          vrchol.Tag = new InformaceVrcholNejkratsiCesta(double.MaxValue);
        }
      }

      try
      {
        // Ohodnocování vrcholů-------------------------------------------------------------
        OhodnotVrcholyNejkratsiCesta(pocatecniVrchol, pocatecniVrchol, koncovyVrchol);

        // Hledání nejkratší cesty----------------------------------------------------------
        return NajdiCestuNejkratsiCesta(koncovyVrchol, pocatecniVrchol);
      }
      catch
      {
        throw new ArgumentException("Graf není obyčejný nebo není na dané cestě souvislý");
      }
    }
    #endregion // Privátní

    #region Veřejné
    #region Nejkratší cesta
    /// <summary>
    /// Metoda se pokusí nalézt nejkratší cestu na grafu dle zadaných parametrů.
    /// <remarks>
    /// Metoda využívá vlastnosti Tag u hran a vrcholů. Ohodnocení hran musí být v intervalu (0 ; double.MaxValue).
    /// </remarks>
    /// </summary>
    /// <param name="graf">Graf, na kterém se má hledat.</param>
    /// <param name="pocatecniVrchol">Počátečný vrchol cesty.</param>
    /// <param name="koncovyVrchol">Koncový vrchol cesty.</param>
    /// <param name="extraktor">Metoda pro extrahování číselné hodnoty ohodnocení hrany z datového typu pro ohodnocení.</param>
    /// <returns>Seznam prvků grafu tvořících nejkratší cestu (prvky jdou od prvního dle cesty) 
    /// Tagem koncového vrcholu je číslo obsahující délku cesty.</returns>
    /// <exception cref="System.ArgumentException">Metoda vyvolá výjimku, pokud graf není obyčejný, hodnoty hran nejsou v intervalu
    /// (0 ; double.MaxValue) nebo pokud graf není souvislý na hledané cestě.</exception>
    public static AdsObousmernySeznam<IPrvekGrafu> NejkratsiCesta(
      AdsGrafSmisenaReprezentace<TVrchol, THrana, TKlicVrcholu, TKlicHrany, TOhodnoceniHrany> graf,
      TVrchol pocatecniVrchol,
      TVrchol koncovyVrchol,
      ExtrahujHodnotuOhodnoceniCallback<TOhodnoceniHrany> extraktor)
    {
      // Počáteční ohodnocení tagů hran
      foreach (IPrvekTabulky<THrana, TOhodnoceniHrany> item in graf.SeznamHran)
      {
        THrana hrana = item.Data;
        double ohodnoceni = extraktor((item.Data as THrana).Ohodnoceni);
        if (ohodnoceni > 0 && ohodnoceni < double.MaxValue)
        {
          hrana.Tag = new InformaceHranaNejkratsiCesta(ohodnoceni);
        }
        else
        {
          throw new ArgumentException("Hrana " + hrana.ToString() + " nesplňuje kritéria ohodnocení");
        }
      }

      AdsObousmernySeznam<IPrvekGrafu> cesta = _NejkratsiCesta(graf, pocatecniVrchol,
        koncovyVrchol);
      // Hodnota cesty
      (cesta.ZpristupniPosledni() as TVrchol).Tag = (
        (
          (
            cesta.ZpristupniPosledni() as TVrchol
          ).Tag as InformaceVrcholNejkratsiCesta
        ).Ohodnceni
      );

      return cesta;
    }
    #endregion // Nejkratší cesta

    #region Nejspolehlivější cesta
    /// <summary>
    /// Metoda se pokusí nalézt nejspolehlivější cestu na grafu dle zadaných parametrů.
    /// <remarks>
    /// Metoda využívá vlastnosti Tag u hran a vrcholů. Ohodnocení hran musí být z intervalu (0 ; 1) včetně.
    /// </remarks>
    /// </summary>
    /// <param name="graf">Graf, na kterém se má hledat.</param>
    /// <param name="pocatecniVrchol">Počátečný vrchol cesty.</param>
    /// <param name="koncovyVrchol">Koncový vrchol cesty.</param>
    /// <param name="extraktor">Metoda pro extrahování číselné hodnoty ohodnocení hrany z datového typu pro ohodnocení.</param>
    /// <returns>Seznam prvků grafu tvořících nejspolehlivější cestu (prvky jdou od prvního dle cesty)
    /// Tagem koncového vrcholu je číslo obsahující délku cesty.</returns>
    /// <exception cref="System.ArgumentException">Metoda vyvolá výjimku, pokud graf není obyčejný,
    /// nebo pokud graf není souvislý na hledané cestě.</exception>
    public static AdsObousmernySeznam<IPrvekGrafu> NejspolehlivejsiCesta(
    AdsGrafSmisenaReprezentace<TVrchol, THrana, TKlicVrcholu, TKlicHrany, TOhodnoceniHrany> graf,
    TVrchol pocatecniVrchol,
    TVrchol koncovyVrchol,
      ExtrahujHodnotuOhodnoceniCallback<TOhodnoceniHrany> extraktor
    )
    {
      // Přeznačení hran
      foreach (IPrvekTabulky<THrana, TOhodnoceniHrany> item in graf.SeznamHran)
      {
        THrana hrana = item.Data;
        double ohodnoceni = extraktor((item.Data as THrana).Ohodnoceni);

        if (ohodnoceni >= 0 && ohodnoceni <= 1)
        {
          if (ohodnoceni == 0)
          {
            hrana.Tag = new InformaceHranaNejkratsiCesta(Math.Round(-Math.Log(0.001, 10), 6));
          }
          else if (ohodnoceni == 1)
          {
            hrana.Tag = new InformaceHranaNejkratsiCesta(Math.Round(-Math.Log(0.999, 10), 6));
          }
          else
          {
            hrana.Tag = new InformaceHranaNejkratsiCesta(Math.Round(-Math.Log(ohodnoceni, 10), 6));
          }
        }
        else
        {
          throw new ArgumentException("Hrana " + hrana.ToString() + " nesplňuje kritéria ohodnocení");
        }

      }

      AdsObousmernySeznam<IPrvekGrafu> cesta = _NejkratsiCesta(graf, pocatecniVrchol,
        koncovyVrchol);

      double hodnotaCesty = 0;
      bool prvniPrirazeni = true;
      foreach (IPrvekGrafu prvek in cesta)
      {
        if (prvek is THrana)
        {
          if (prvniPrirazeni)
          {
            hodnotaCesty = extraktor((prvek as THrana).Ohodnoceni);
            prvniPrirazeni = false;
          }
          else
          {
            hodnotaCesty = hodnotaCesty * extraktor((prvek as THrana).Ohodnoceni);
          }
        }
      }

      (cesta.ZpristupniPosledni() as TVrchol).Tag = hodnotaCesty;

      return cesta;
    }
    #endregion // Nejspolehlivější cesta
    #endregion // Veřejné
    #endregion // Metody
  }
}